# -*- coding: utf-8 -*-

import pandas as pd
from pandas.api.types import is_numeric_dtype

from copy import copy

from .trajectory import Trajectory
from .trajectory_collection import TrajectoryCollection
from .unit_utils import UNITS, get_conversion
from .spatiotemporal_utils import TPoint, get_speed


class TrajectoryCleaner:
    """
    Cleaner base class
    """

    def __init__(self, traj):
        """
        Create TrajectoryCleaner

        Parameters
        ----------
        traj : Trajectory or TrajectoryCollection
        """
        self.traj = traj

    def clean(self, **kwargs):
        """
        Clean the input Trajectory/TrajectoryCollection.

        Returns
        -------
        Trajectory/TrajectoryCollection
            Cleaned Trajectory or TrajectoryCollection
        """
        if isinstance(self.traj, Trajectory):
            return self._clean_traj(self.traj, **kwargs)
        elif isinstance(self.traj, TrajectoryCollection):
            return self._clean_traj_collection(**kwargs)
        else:
            raise TypeError

    def _clean_traj_collection(self, **kwargs):
        cleaned = []
        for traj in self.traj:
            cleaned.append(self._clean_traj(traj, **kwargs))
        result = copy(self.traj)
        result.trajectories = cleaned
        return result

    def _clean_traj(self, traj, **kwargs):
        return traj


class IqrCleaner(TrajectoryCleaner):
    """
    Interquantile range (IQR) based outlier cleaner.

    Parameters
    ----------
    columns: dictionary
        Information regarding the columns that will be used to clean the trajectory
        and the accompanying thresholds.

        Key-value pairs of columns and alpha (IQR multiplier).
        Note: Setting alpha=3 is widely used.

    Examples
    --------

    >>> mpd.IqrCleaner(traj).clean(columns={'speed': 3})
    """

    def _clean_traj(self, traj, columns):
        df = traj.df.copy()
        ixs = []

        for column, alpha in columns.items():
            if not is_numeric_dtype(df[column]):
                raise TypeError(
                    f"'{column}' column of type '{df[column].dtype}' is not numeric"
                )
            ix = self._calc_outliers(df[column], alpha)
            ixs.append(ix.tolist())

        indices = pd.Series(list(map(any, zip(*ixs))), index=df.index)
        return Trajectory(df[~indices], traj.id, traj_id_col=traj.get_traj_id_col())

    def _calc_outliers(self, series, alpha=3):
        """
        Returns a series of indexes of row that are to be considered outliers
        using the quantiles of the data.
        """
        q25, q75 = series.quantile((0.25, 0.75))
        iqr = q75 - q25
        q_high = q75 + alpha * iqr
        q_low = q25 - alpha * iqr
        # return the row indexes that are over/under the calculated threshold
        return (series > q_high) | (series < q_low)


class OutlierCleaner(TrajectoryCleaner):
    """
    Speed-based outlier cleaner that cuts away spikes in the trajectory
    when the speed exceeds the provided speed threshold value

    Parameters
    ----------
    v_max : numeric
        Speed threshold
    units : tuple
        Units in which to calculate speed
        For more info, check the list of supported units at
        https://movingpandas.org/units

    Examples
    --------

    >>> mpd.OutlierCleaner(traj).clean(alpha=2)

    >>> mpd.OutlierCleaner(traj).clean(v_max=100, units=("km", "h"))
    """

    def _clean_traj(self, traj, v_max=None, units=UNITS(), alpha=3):
        out_traj = traj.copy()
        conversion = get_conversion(units, traj.crs_units)

        if v_max is None:
            out_traj.add_speed(overwrite=True, units=units)
            speed_col = out_traj.get_speed_col()
            v_max = out_traj.df[speed_col].agg(lambda x: x.quantile(0.95))
            v_max = v_max * alpha

        ixs = []
        prev = None
        for index, row in out_traj.df.iterrows():
            curr = TPoint(index, row.geometry)
            if not prev:
                prev = curr
                continue

            v = get_speed(prev, curr, traj.is_latlon, conversion)
            if v > v_max:
                ixs.append(index)
                continue  # do NOT update the previous point
            prev = curr

        out_traj.df.drop(ixs, inplace=True)
        return out_traj



# -*- coding: utf-8 -*-

import pytest
from datetime import datetime, timedelta
from pytz import timezone

from pyproj import CRS
from numpy import issubdtype

from movingpandas.trajectory_collection import TrajectoryCollection
from movingpandas.trajectory_splitter import StopSplitter
from movingpandas.trajectory_stop_detector import TrajectoryStopDetector
from .test_trajectory import Node, make_traj

CRS_METRIC = CRS.from_user_input(31256)
CRS_LATLON = CRS.from_user_input(4326)


class TestTrajectoryStopDetector:
    def setup_method(self):
        self.traj = make_traj(
            [
                Node(0, 0),
                Node(0, 10, second=1),
                Node(0, 20, second=2),
                Node(0, 21, second=4),
                Node(0, 22, second=6),
                Node(0, 30, second=8),
                Node(0, 40, second=10),
                Node(1, 50, second=15),
            ]
        )
        self.detector = TrajectoryStopDetector(self.traj)

    def test_stop_segments_middle_stop(self):
        stop_segments = self.detector.get_stop_segments(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        assert len(stop_segments) == 1
        assert (
            stop_segments.trajectories[0].to_linestringm_wkt()
            == "LINESTRING M (0.0 20.0 2.0, 0.0 21.0 4.0, 0.0 22.0 6.0)"
        )
        assert stop_segments.get_crs() == CRS_METRIC

    def test_stop_times_middle_stop(self):
        stop_times = self.detector.get_stop_time_ranges(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        assert len(stop_times) == 1
        assert stop_times[0].t_0 == datetime(1970, 1, 1, 0, 0, 2)
        assert stop_times[0].t_n == datetime(1970, 1, 1, 0, 0, 6)

    def test_stop_points_middle_stop(self):
        self.traj.id = "a"
        stop_points = self.detector.get_stop_points(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        assert len(stop_points) == 1
        assert stop_points.iloc[0].geometry.x == 0
        assert stop_points.iloc[0].geometry.y == 21
        assert stop_points.iloc[0].start_time == datetime(1970, 1, 1, 0, 0, 2)
        assert stop_points.iloc[0].end_time == datetime(1970, 1, 1, 0, 0, 6)
        assert stop_points.iloc[0].duration_s == 4
        assert stop_points.iloc[0].traj_id == "a"

    def test_stop_point_parent_traj_id(self):
        stop_points = self.detector.get_stop_points(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        assert stop_points.iloc[0].traj_id == 1
        assert issubdtype(stop_points.iloc[0].traj_id, type(self.traj.id))
        self.traj.id = "a"
        stop_points = self.detector.get_stop_points(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        assert stop_points.iloc[0].traj_id == "a"
        assert isinstance(stop_points.iloc[0].traj_id, type(self.traj.id))
        self.traj.id = 5.5
        stop_points = self.detector.get_stop_points(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        assert stop_points.iloc[0].traj_id == 5.5
        assert issubdtype(stop_points.iloc[0].traj_id, type(self.traj.id))

    def test_stop_detector_start_stop(self):
        traj = make_traj(
            [
                Node(0, 0),
                Node(0, 1, second=1),
                Node(0, 2, second=2),
                Node(0, 1, second=3),
                Node(0, 22, second=4),
                Node(0, 30, second=8),
                Node(0, 40, second=10),
                Node(1, 50, second=15),
            ]
        )
        detector = TrajectoryStopDetector(traj)
        stop_times = detector.get_stop_time_ranges(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        assert len(stop_times) == 1
        assert stop_times[0].t_0 == datetime(1970, 1, 1, 0, 0, 0)
        assert stop_times[0].t_n == datetime(1970, 1, 1, 0, 0, 3)

    def test_stop_detector_end_stop(self):
        traj = make_traj(
            [
                Node(0, -100),
                Node(0, -10, second=1),
                Node(0, 2, second=2),
                Node(0, 1, second=3),
                Node(0, 22, second=4),
                Node(0, 30, second=8),
                Node(0, 31, second=10),
                Node(1, 32, second=15),
            ]
        )
        detector = TrajectoryStopDetector(traj)
        stop_times = detector.get_stop_time_ranges(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        assert len(stop_times) == 1
        assert stop_times[0].t_0 == datetime(1970, 1, 1, 0, 0, 8)
        assert stop_times[0].t_n == datetime(1970, 1, 1, 0, 0, 15)

    def test_stop_detector_multiple_stops(self):
        traj = make_traj(
            [
                Node(0, 0),
                Node(0, 1, second=1),
                Node(0, 2, second=2),
                Node(0, 1, second=3),
                Node(0, 22, second=4),
                Node(0, 30, second=8),
                Node(0, 31, second=10),
                Node(1, 32, second=15),
            ]
        )
        detector = TrajectoryStopDetector(traj)
        stop_times = detector.get_stop_time_ranges(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        assert len(stop_times) == 2
        assert stop_times[0].t_0 == datetime(1970, 1, 1, 0, 0, 0)
        assert stop_times[0].t_n == datetime(1970, 1, 1, 0, 0, 3)
        assert stop_times[1].t_0 == datetime(1970, 1, 1, 0, 0, 8)
        assert stop_times[1].t_n == datetime(1970, 1, 1, 0, 0, 15)

    def test_stop_detector_multiple_stops_without_gap(self):
        traj = make_traj(
            [
                Node(0, 0),
                Node(0, 1, second=1),
                Node(0, 2, second=2),
                Node(0, 1, second=3),
                Node(0, 30, second=8),
                Node(0, 31, second=10),
                Node(1, 32, second=15),
            ]
        )
        detector = TrajectoryStopDetector(traj)
        stop_times = detector.get_stop_time_ranges(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        assert len(stop_times) == 2
        assert stop_times[0].t_0 == datetime(1970, 1, 1, 0, 0, 0)
        assert stop_times[0].t_n == datetime(1970, 1, 1, 0, 0, 3)
        assert stop_times[1].t_0 == datetime(1970, 1, 1, 0, 0, 8)
        assert stop_times[1].t_n == datetime(1970, 1, 1, 0, 0, 15)

    def test_stop_detector_no_stops(self):
        traj = make_traj(
            [
                Node(0, 0),
                Node(0, 1, second=1),
                Node(0, 2, second=2),
                Node(0, 1, second=3),
                Node(0, 22, second=4),
                Node(0, 30, second=8),
                Node(0, 31, second=10),
                Node(1, 32, second=15),
            ]
        )
        detector = TrajectoryStopDetector(traj)
        stop_times = detector.get_stop_time_ranges(
            max_diameter=1, min_duration=timedelta(seconds=1)
        )
        stop_segments = detector.get_stop_segments(
            max_diameter=1, min_duration=timedelta(seconds=1)
        )
        stop_points = detector.get_stop_points(
            max_diameter=1, min_duration=timedelta(seconds=1)
        )
        assert len(stop_times) == 0
        assert len(stop_segments) == 0
        assert len(stop_points) == 0

    def test_stop_detector_tzaware_data(self):
        nodes = [
            Node(0, 0),
            Node(0, 1, second=1),
            Node(0, 2, second=2),
            Node(0, 1, second=3),
            Node(0, 22, second=4),
            Node(0, 30, second=8),
            Node(0, 31, second=10),
            Node(1, 32, second=15),
        ]

        for node in nodes:
            node.t = timezone("Europe/Vienna").localize(node.t)

        with pytest.warns():
            traj = make_traj(nodes)
        detector = TrajectoryStopDetector(traj)
        stop_times = detector.get_stop_time_ranges(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        assert len(stop_times) == 2
        assert stop_times[0].t_0 == datetime(1970, 1, 1, 0, 0, 0)
        assert stop_times[0].t_n == datetime(1970, 1, 1, 0, 0, 3)
        assert stop_times[1].t_0 == datetime(1970, 1, 1, 0, 0, 8)
        assert stop_times[1].t_n == datetime(1970, 1, 1, 0, 0, 15)
        assert traj.df.index.tzinfo is None

    def test_stop_detector_collection(self):
        traj1 = make_traj(
            [
                Node(0, 0),
                Node(0, 1, second=1),
                Node(0, 2, second=2),
                Node(0, 1, second=3),
                Node(0, 22, second=4),
                Node(0, 30, second=8),
                Node(0, 40, second=10),
                Node(1, 50, second=15),
            ],
            id=1,
        )
        traj2 = make_traj(
            [
                Node(0, -100),
                Node(0, -10, second=1),
                Node(0, 2, second=2),
                Node(0, 1, second=3),
                Node(0, 22, second=4),
                Node(0, 30, second=8),
                Node(0, 31, second=10),
                Node(1, 32, second=15),
            ],
            id=2,
        )
        collection = TrajectoryCollection([traj1, traj2])
        detector = TrajectoryStopDetector(collection)
        stop_times = detector.get_stop_time_ranges(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        stop_segments = detector.get_stop_segments(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        stop_points = detector.get_stop_points(
            max_diameter=3, min_duration=timedelta(seconds=2)
        )
        assert len(stop_times) == 2
        assert len(stop_segments) == 2
        assert len(stop_points) == 2

    def test_stop_splitter_no_stops(self):
        traj1 = make_traj(
            [
                Node(0, 0),
                Node(0, 10, second=10),
                Node(0, 20, second=20),
                Node(0, 30, second=30),
                Node(0, 40, second=40),
                Node(0, 50, second=50),
            ],
            id=1,
        )
        collection = TrajectoryCollection([traj1])
        detector = StopSplitter(collection)
        stops = detector.split(max_diameter=1, min_duration=timedelta(seconds=1))
        assert len(stops) == 1



from copy import copy
from abc import ABC, abstractmethod

import numpy as np
from shapely.geometry import Point

try:
    from stonesoup.types.detection import Detection
    from stonesoup.models.transition.linear import (
        CombinedLinearGaussianTransitionModel,
        ConstantVelocity,
    )
    from stonesoup.models.measurement.linear import LinearGaussian
    from stonesoup.predictor.kalman import KalmanPredictor
    from stonesoup.updater.kalman import KalmanUpdater
    from stonesoup.types.array import CovarianceMatrix, StateVector
    from stonesoup.types.prediction import GaussianStatePrediction
    from stonesoup.initiator.simple import SimpleMeasurementInitiator
    from stonesoup.types.hypothesis import SingleHypothesis
    from stonesoup.smoother.kalman import KalmanSmoother
    from stonesoup.reader.base import DetectionReader
    from stonesoup.base import Property
    from stonesoup.buffered_generator import BufferedGenerator
except ImportError as error:
    raise ImportError(
        "Missing optional dependencies. To use the trajectory smoother classes please "
        "install Stone Soup "
        "(see https://stonesoup.readthedocs.io/en/latest/#installation)."
    ) from error

from .trajectory import Trajectory
from .trajectory_collection import TrajectoryCollection


class TrajectorySmoother(ABC):
    """
    TrajectorySmoother base class

    Base class for trajectory smoothers. This class is abstract and thus cannot be
    instantiated.
    """

    def __init__(self, traj):
        """
        Create TrajectorySmoother

        Parameters
        ----------
        traj : Trajectory or TrajectoryCollection
        """
        self.traj = traj

    def smooth(self, **kwargs):
        """
        Smooth the input Trajectory/TrajectoryCollection

        Parameters
        ----------
        kwargs : any type
            Keyword arguments, differs by smoother

        Returns
        -------
        Trajectory/TrajectoryCollection
            Smoothed Trajectory or TrajectoryCollection

        """
        if isinstance(self.traj, Trajectory):
            return self._smooth_traj(self.traj, **kwargs)
        elif isinstance(self.traj, TrajectoryCollection):
            return self._smooth_traj_collection(**kwargs)
        else:
            raise TypeError

    def _smooth_traj_collection(self, **kwargs):
        smoothed = []
        for traj in self.traj:
            smoothed.append(self._smooth_traj(traj, **kwargs))
        result = copy(self.traj)
        result.trajectories = smoothed
        return result

    @abstractmethod
    def _smooth_traj(self, traj, **kwargs):
        raise NotImplementedError


class KalmanSmootherCV(TrajectorySmoother):
    """
    Smooths using a Kalman Filter with a Constant Velocity model.

    The Constant Velocity model assumes that the speed between consecutive locations is
    nearly constant. For trajectories where ``traj.is_latlon = True`` the smoother
    converts to EPSG:3395 (World Mercator) internally to perform filtering and smoothing

    .. note::
        This class makes use of
        `Stone Soup <https://stonesoup.readthedocs.io/en/latest/>`_, which is an
        optional dependency and not installed by default. To use this class, you need
        to install Stone Soup directly
        (see `here <https://stonesoup.readthedocs.io/en/latest/#installation>`_).
    """

    def smooth(self, process_noise_std=0.5, measurement_noise_std=1):
        """
        Smooth the input Trajectory/TrajectoryCollection

        Parameters
        ----------
        process_noise_std: float or sequence of floats of length 2, default is 0.5.
            The process (acceleration) noise standard deviation.

            If a sequence (e.g. list, tuple, etc.) is provided, the first index is used
            for the x coordinate, while the second is used for the y coordinate. If
            ``traj.is_latlon=True`` the values are applied to the  easting and northing
            coordinate (in EPSG:3395) respectively.

            Alternatively, a single float can be provided, which is assumed to be the
            same for both coordinates.

            This governs the uncertainty associated with the adherence of the new
            (smooth) trajectories to the CV model assumption; higher values relax
            the assumption, therefore leading to less-smooth trajectories,
            and vice-versa.

        measurement_noise_std: float or sequence of floats of length 2, default is 1.
            The measurement noise standard deviation.

            If a sequence (e.g. list, tuple, etc.) is provided, the first index is used
            for the x coordinate, while the second is used for the y coordinate. If
            ``traj.is_latlon=True`` the values are applied to the  easting and northing
            coordinate (in EPSG:3395) respectively.

            Alternatively, a single float can be provided, which is assumed to be the
            same for both coordinates.

            This controls the assumed error in the original trajectories; higher values
            dictate that the original trajectories are expected to be noisier
            (and therefore, less reliable), thus leading to smoother trajectories,
            and vice-versa.
        """
        return super().smooth(
            process_noise_std=process_noise_std,
            measurement_noise_std=measurement_noise_std,
        )

    def _smooth_traj(self, traj, process_noise_std=0.5, measurement_noise_std=1):
        # Get detector
        detector = self._get_detector(traj)

        # Models
        if not isinstance(process_noise_std, (list, tuple, np.ndarray)):
            process_noise_std = [process_noise_std, process_noise_std]
        if not isinstance(measurement_noise_std, (list, tuple, np.ndarray)):
            measurement_noise_std = [measurement_noise_std, measurement_noise_std]
        transition_model = CombinedLinearGaussianTransitionModel(
            [
                ConstantVelocity(process_noise_std[0] ** 2),
                ConstantVelocity(process_noise_std[1] ** 2),
            ]
        )
        measurement_model = LinearGaussian(
            ndim_state=4,
            mapping=[0, 2],
            noise_covar=np.diag(
                [measurement_noise_std[0] ** 2, measurement_noise_std[1] ** 2]
            ),
        )
        # Predictor and updater
        predictor = KalmanPredictor(transition_model)
        updater = KalmanUpdater(measurement_model)
        # Initiator
        state_vector = StateVector([0.0, 0.0, 0.0, 0.0])
        covar = CovarianceMatrix(np.diag([0.0, 0.0, 0.0, 0.0]))
        prior_state = GaussianStatePrediction(state_vector, covar)
        initiator = SimpleMeasurementInitiator(prior_state, measurement_model)
        # Filtering
        track = None
        for i, (timestamp, detections) in enumerate(detector):
            if i == 0:
                tracks = initiator.initiate(detections, timestamp)
                track = tracks.pop()
            else:
                detection = detections.pop()
                prediction = predictor.predict(track.state, timestamp=timestamp)
                hypothesis = SingleHypothesis(prediction, detection)
                posterior = updater.update(hypothesis)
                track.append(posterior)
        # Smoothing
        smoother = KalmanSmoother(transition_model)
        smooth_track = smoother.smooth(track)

        # Create new trajectory
        if traj.is_latlon:
            df = traj.df.to_crs("EPSG:3395")
            df.geometry = [
                Point(state.state_vector[0], state.state_vector[2])
                for state in smooth_track
            ]
            df.to_crs(traj.crs, inplace=True)
        else:
            df = traj.df.copy()
            df.geometry = [
                Point(state.state_vector[0], state.state_vector[2])
                for state in smooth_track
            ]
        new_traj = Trajectory(df, traj.id, traj_id_col=traj.get_traj_id_col())
        return new_traj

    @staticmethod
    def _get_detector(traj):
        class Detector(DetectionReader):
            trajectory: Trajectory = Property(doc="")

            @BufferedGenerator.generator_method
            def detections_gen(self):
                detections_df = traj.df.loc[:, "geometry"].to_frame()
                if traj.is_latlon:
                    detections_df.to_crs("EPSG:3395", inplace=True)
                detections_df["x"] = [
                    row.geometry.coords[0][0] for _, row in detections_df.iterrows()
                ]
                detections_df["y"] = [
                    row.geometry.coords[0][1] for _, row in detections_df.iterrows()
                ]
                detections_df.drop(columns="geometry", inplace=True)
                for time, row in detections_df.iterrows():
                    detection = Detection([row.x, row.y], timestamp=time)
                    yield time, {detection}

        return Detector(traj)



# -*- coding: utf-8 -*-

from .trajectory import Trajectory
from .trajectory_collection import TrajectoryCollection


def convert_time_ranges_to_segments(traj, time_ranges):
    """
    Extracts trajectory segments for the specified time ranges.

    Parameters
    ----------
    traj : Trajectory/TrajectoryCollection
        Trajectory or TrajectoryCollection from which to extract the time_ranges
    time_ranges : list
        List of TimeRangeWithTrajId objects

    Returns
    -------
    list
        Extracted trajectory segements
    """
    if isinstance(traj, Trajectory):
        handle_collection = False
    elif isinstance(traj, TrajectoryCollection):
        handle_collection = True
    else:
        raise TypeError

    segments = []
    for time_range in time_ranges:
        try:
            if handle_collection:
                segments.append(
                    traj.get_trajectory(time_range.traj_id).get_segment_between(
                        time_range.t_0, time_range.t_n
                    )
                )
            else:
                segments.append(
                    traj.get_segment_between(time_range.t_0, time_range.t_n)
                )
        except ValueError:
            pass
    return segments



# -*- coding: utf-8 -*-

import warnings

from shapely.affinity import translate
from shapely.geometry import Point, LineString
from pandas import DataFrame, to_datetime, Series
from pandas.core.indexes.datetimes import DatetimeIndex
from geopandas import GeoDataFrame

try:
    from pyproj import CRS
except ImportError:
    # TODO: fiona.crs is deprecated from fiona 2.0, use fiona.CRS instead
    from fiona.crs import from_epsg

from .overlay import clip, intersection, intersects, create_entry_and_exit_points
from .spatiotemporal_utils import STRange
from .geometry_utils import (
    angular_difference,
    azimuth,
    calculate_initial_compass_bearing,
    measure_distance,
    measure_distance_line,
    measure_length,
    point_gdf_to_linestring,
)
from .unit_utils import (
    UNITS,
    MissingCRSWarning,
    to_unixtime,
    get_conversion,
)
from .spatiotemporal_utils import get_speed2
from .trajectory_plotter import _TrajectoryPlotter
from .io import gdf_to_mf_json

warnings.filterwarnings(  # see https://github.com/movingpandas/movingpandas/issues/289
    "ignore", message="CRS not set for some of the concatenation inputs."
)

ACCELERATION_COL_NAME = "acceleration"
ANGULAR_DIFFERENCE_COL_NAME = "angular_difference"
DIRECTION_COL_NAME = "direction"
DISTANCE_COL_NAME = "distance"
SPEED_COL_NAME = "speed"
TIMEDELTA_COL_NAME = "timedelta"
TRAJ_ID_COL_NAME = "traj_id"


class TimeZoneWarning(UserWarning, ValueError):
    pass


class Trajectory:
    def __init__(
        self,
        df,
        traj_id,
        traj_id_col=None,
        obj_id=None,
        t=None,
        x=None,
        y=None,
        crs="epsg:4326",
        parent=None,
    ):
        """
        Create Trajectory from GeoDataFrame or DataFrame.

        Parameters
        ----------
        df : GeoDataFrame or DataFrame
            GeoDataFrame with point geometry column and timestamp index
        traj_id : any
            Trajectory ID
        obj_id : any
            Moving object ID
        t : string
            Name of the DataFrame column containing the timestamp
        x : string
            Name of the DataFrame column containing the x coordinate
        y : string
            Name of the DataFrame column containing the y coordinate
        crs : string
            CRS of the x/y coordinates
        parent : Trajectory
            Parent trajectory

        Examples
        --------
        Creating a trajectory from scratch:

        >>> import pandas as pd
        >>> import geopandas as gpd
        >>> import movingpandas as mpd
        >>>
        >>> df = pd.DataFrame([
        ...     {'geometry':Point(0,0), 't':datetime(2018,1,1,12,0,0)},
        ...     {'geometry':Point(6,0), 't':datetime(2018,1,1,12,6,0)},
        ...     {'geometry':Point(6,6), 't':datetime(2018,1,1,12,10,0)},
        ...     {'geometry':Point(9,9), 't':datetime(2018,1,1,12,15,0)}
        ... ]).set_index('t')
        >>> gdf = gpd.GeoDataFrame(df, crs=31256)
        >>> traj = mpd.Trajectory(gdf, 1)

        For more examples, see the tutorial notebooks_.

        .. _notebooks: https://mybinder.org/v2/gh/movingpandas/movingpandas/main?filepath=tutorials/1-getting-started.ipynb
        """  # noqa: E501

        if len(df) < 2:
            raise ValueError("The input DataFrame must have at least two rows.")
        if not isinstance(df, GeoDataFrame):
            if x is None or y is None:
                raise ValueError(
                    "The input DataFrame needs to be a GeoDataFrame or x and y columns"
                    "need to be specified."
                )
            df = GeoDataFrame(
                df.drop([x, y], axis=1),
                crs=crs,
                geometry=[Point(xy) for xy in zip(df[x], df[y])],
            )
        if not isinstance(df.index, DatetimeIndex):
            if t is None:
                raise TypeError(
                    "The input DataFrame needs a DatetimeIndex or a timestamp column"
                    "needs to be specified. Use Pandas' set_index() method to create an"
                    "index or specify the timestamp column name."
                )
            df[t] = to_datetime(df[t])
            df = df.set_index(t)

        #  Drop any time zone information, to avoid errors (see issue 303)
        if df.index.tzinfo is not None:
            self.df_orig_tz = df.index.tzinfo
            warnings.warn(
                "Time zone information dropped from trajectory. "
                "All dates and times will use local time. "
                "This is applied by doing df.tz_localize(None). "
                "To use UTC or a different time zone, convert and drop "
                "time zone information prior to trajectory creation.",
                category=TimeZoneWarning,
            )
            df = df.tz_localize(None)

        self.id = traj_id
        self.obj_id = obj_id
        df.sort_index(inplace=True)
        self.df = df[~df.index.duplicated(keep="first")].copy()
        self.crs = df.crs
        self.parent = parent

        if traj_id_col is not None:
            self.traj_id_col_name = traj_id_col
        else:
            self.traj_id_col_name = TRAJ_ID_COL_NAME
        self.df[self.traj_id_col_name] = traj_id
        self.df[self.traj_id_col_name].astype("category")

        if self.crs is not None:
            self.crs_units = self.crs.axis_info[0].unit_name
        else:
            self.crs_units = None
            warnings.warn(
                "Trajectory generated without CRS. Computations will use Euclidean"
                " distances.",
                category=MissingCRSWarning,
            )
            self.is_latlon = False
            return
        try:
            crs = CRS.from_user_input(self.crs)
            self.is_latlon = crs.is_geographic
        except NameError:
            self.is_latlon = self.crs["init"] == from_epsg(4326)["init"]

    def __str__(self):
        try:
            line = self.to_linestring()
        except RuntimeError:
            return "Invalid trajectory!"
        return (
            f"Trajectory {self.id} ({self.get_start_time()} to {self.get_end_time()}) "
            f"| Size: {self.size()} | Length: {round(self.get_length(), 1)}m\n"
            f"Bounds: {self.get_bbox()}\n{line.wkt[:100]}"
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        # TODO: make bullet proof
        return (
            str(self) == str(other)
            and self.crs == other.crs
            and self.parent == other.parent
        )

    def size(self):
        """
        Returns number of rows in Trajectory.df

        Returns
        -------
        size : int
            Number of rows
        """
        return len(self.df.index)

    def copy(self):
        """
        Return a copy of the trajectory.

        Returns
        -------
        Trajectory
        """
        copied = Trajectory(
            self.df.copy(),
            self.id,
            parent=self.parent,
            traj_id_col=self.traj_id_col_name,
        )
        return copied

    def drop(self, **kwargs):
        """
        Drop columns or rows from the trajectory DataFrame

        Examples
        --------

        >>> trajectory.drop(columns=['abc','def'])
        """
        self.df.drop(**kwargs, inplace=True)

    def plot(self, *args, **kwargs):
        """
        Generate a plot using GeoPandas default plotting (Matplotlib).

        Parameters
        ----------
        args :
            These parameters will be passed to the TrajectoryPlotter
        kwargs :
            These parameters will be passed to the TrajectoryPlotter

        Returns
        -------
        Matplotlib plot

        Examples
        --------
        Plot speed along trajectory (with legend and specified figure size):

        >>> trajectory.plot(column='speed', legend=True, figsize=(9,5))
        """
        return _TrajectoryPlotter(self, *args, **kwargs).plot()

    def explore(self, *args, **kwargs):
        """
        Generate a plot using GeoPandas explore (folium/leaflet.js)
        https://geopandas.org/en/stable/docs/reference/api/geopandas.GeoDataFrame.explore.html

        Parameters
        ----------
        args :
            These parameters will be passed to GeoPandas explore
        kwargs :
            These parameters will be passed to GeoPandas explore

        Returns
        -------
        m : folium.folium.Map
            folium Map instance

        Examples
        --------
        Plot speed along trajectory (with legend and specified figure size):

        >>> trajectory.explore(column='speed', vmax=20, tiles="CartoDB positron")
        """  # noqa: E501
        from importlib.metadata import version

        if version("geopandas") >= "1.0.0":
            return self.to_line_gdf().explore(*args, **kwargs)
        else:
            raise (
                NotImplementedError(
                    "Please install geopandas >= 1.0.0 to use this function."
                )
            )

    def hvplot(self, *args, **kwargs):
        """
        Generate an interactive plot using HoloViews.

        The following parameters are set by default:

        geo=True, tiles='OSM', marker_size=200, line_width=2.0

        Parameters
        ----------
        args :
            These parameters will be passed to the TrajectoryPlotter
        kwargs :
            These parameters will be passed to the TrajectoryPlotter

        Returns
        -------
        Holoviews plot

        Examples
        --------
        Plot speed along trajectory (with legend and specified figure size):

        >>> trajectory.hvplot(c='speed', line_width=7.0, width=700, height=400, colorbar=True)
        """  # noqa: E501
        return _TrajectoryPlotter(self, *args, **kwargs).hvplot()

    def hvplot_pts(self, *args, **kwargs):
        """
        Generate an interactive plot of trajectory points.

        Parameters
        ----------
        args :
            These parameters will be passed to the TrajectoryPlotter
        kwargs :
            These parameters will be passed to the TrajectoryPlotter

            To customize the plots, check the list of supported colormaps:
            https://holoviews.org/user_guide/Colormaps.html#available-colormaps

        Examples
        --------
        Plot points colored by speed (with legend and specified figure size):

        >>> collection.hvplot_pts(c='speed', width=700, height=400, colorbar=True)
        """  # noqa: E501
        return _TrajectoryPlotter(self, *args, **kwargs).hvplot_pts()

    def is_valid(self):
        """
        Return whether the trajectory meets minimum requirements.

        Returns
        -------
        bool
        """
        if len(self.df) < 2:
            return False
        if not self.get_start_time() < self.get_end_time():
            return False
        return True

    def get_crs(self):
        """
        Return the CRS of the trajectory
        """
        return self.crs

    def to_crs(self, crs):
        """
        Return the trajectory reprojected to the target CRS

        Parameters
        ----------
        crs : pyproj.CRS
            Target coordinate reference system

        Returns
        -------
        Trajectory

        Examples
        --------
        Reproject a trajectory to EPSG:4088

        >>> from pyproj import CRS
        >>> reprojected = trajectory.to_crs(CRS(4088))
        """
        temp = self.copy()
        if type(crs) != CRS:
            crs = CRS(crs)
        temp.crs = crs
        temp.df = temp.df.to_crs(crs)
        temp.is_latlon = crs.is_geographic
        return temp

    def is_latlon(self):
        """
        Return True if the trajectory CRS is geographic (e.g. EPSG:4326 WGS84)
        """
        return self.is_latlon

    def get_min(self, column):
        """
        Return minimum value in the provided DataFrame column

        Parameters
        ----------
        column : string
            Name of the DataFrame column

        Returns
        -------
        Sortable
            Minimum value
        """
        return self.df[column].min()

    def get_max(self, column):
        """
        Return maximum value in the provided DataFrame column

        Parameters
        ----------
        column : string
            Name of the DataFrame column

        Returns
        -------
        Sortable
            Maximum value
        """
        return self.df[column].max()

    def get_column_names(self):
        """
        Return the list of column names

        Returns
        -------
        list
        """
        return self.df.columns

    def get_traj_id_col(self):
        """
        Return name of the trajectory ID column

        Returns
        -------
        string
        """
        if hasattr(self, "traj_id_col_name"):
            return self.traj_id_col_name
        else:
            return TRAJ_ID_COL_NAME

    def get_speed_col(self):
        """
        Return name of the speed column

        Returns
        -------
        string
        """
        if hasattr(self, "speed_col_name"):
            return self.speed_col_name
        else:
            return SPEED_COL_NAME

    def get_distance_col(self):
        """
        Return name of the distance column

        Returns
        -------
        string
        """
        if hasattr(self, "distance_col_name"):
            return self.distance_col_name
        else:
            return DISTANCE_COL_NAME

    def get_direction_col(self):
        """
        Return name of the direction column

        Returns
        -------
        string
        """
        if hasattr(self, "direction_col_name"):
            return self.direction_col_name
        else:
            return DIRECTION_COL_NAME

    def get_angular_difference_col(self):
        """
        Return name of the angular difference column

        Returns
        -------
        string
        """
        if hasattr(self, "angular_difference_col_name"):
            return self.angular_difference_col_name
        else:
            return ANGULAR_DIFFERENCE_COL_NAME

    def get_timedelta_col(self):
        """
        Return name of the timedelta column

        Returns
        -------
        string
        """
        if hasattr(self, "timedelta_col_name"):
            return self.timedelta_col_name
        else:
            return TIMEDELTA_COL_NAME

    def get_geom_col(self):
        """
        Return name of the geometry column

        Returns
        -------
        string
        """
        return self.df.geometry.name

    def to_linestring(self):
        """
        Return trajectory geometry as LineString.

        Returns
        -------
        shapely LineString
        """
        try:
            return point_gdf_to_linestring(self.df, self.get_geom_col())
        except RuntimeError:
            raise RuntimeError("Cannot generate LineString")

    def to_linestringm_wkt(self):
        """
        Return the WKT string of the trajectory LineStringM representation.

        Returns
        -------
        string
            WKT of trajectory as LineStringM
        """
        # Shapely only supports x, y, z. Therefore, this is a bit hacky!
        coords = []
        for index, pt in self.df[self.get_geom_col()].items():
            t = to_unixtime(index)
            coords.append(f"{pt.x} {pt.y} {t}")
        wkt = f"LINESTRING M ({', '.join(coords)})"
        return wkt

    def to_point_gdf(self, return_orig_tz=False):
        """
        Return the trajectory's points as GeoDataFrame.

        Parameters
        ----------
        return_orig_tz : bool
            If True, adds timezone info back to df

        Returns
        -------
        GeoDataFrame
        """
        if return_orig_tz:
            return self.df.tz_localize(self.df_orig_tz)
        return self.df

    def to_line_gdf(self, columns=None):
        """
        Return the trajectory's line segments as GeoDataFrame.

        Returns
        -------
        GeoDataFrame
        """
        line_gdf = self._to_line_df(columns)
        line_gdf.drop(columns=[self.get_geom_col(), "prev_pt"], inplace=True)
        line_gdf.reset_index(drop=True, inplace=True)
        line_gdf.rename(columns={"line": "geometry"}, inplace=True)
        line_gdf.set_geometry("geometry", inplace=True)
        if self.crs:
            line_gdf.set_crs(self.crs, inplace=True)
        return line_gdf

    def to_traj_gdf(self, wkt=False, agg=False):
        """
        Return a GeoDataFrame with one row containing the trajectory as a
        single LineString.

        Parameters
        ----------
        wkt : bool
            If True, adds WKT column representing the trajectory geometry
        agg : dict
            Adds columns with aggregate values computed from trajectory dataframe
            columns according to specified aggregation mode, using
            pandas.DataFrame.agg(), and shortcuts for "mode" and quantiles
            (e.g. "q5" or "q95")

        Examples
        --------

        >>> traj.to_traj_gdf(agg={"col1": "mode", "col2": ["min", "q95"]})

        Returns
        -------
        GeoDataFrame
        """
        properties = {
            self.traj_id_col_name: self.id,
            "start_t": self.get_start_time(),
            "end_t": self.get_end_time(),
            "geometry": self.to_linestring(),
            "length": self.get_length(),
            "direction": self.get_direction(),
        }

        if wkt:
            properties["wkt"] = self.to_linestringm_wkt()
        if agg:
            for col, agg_modes in agg.items():
                if type(agg_modes) != list:
                    agg_modes = [agg_modes]
                for agg_mode in agg_modes:
                    if agg_mode == "mode":
                        aggregated = self.df.agg({col: Series.mode})[col][0]
                    elif agg_mode[0] == "q" and int(agg_mode[1:]) < 100:
                        aggregated = self.df[col].agg(
                            lambda x: x.quantile(int(agg_mode[1:]) / 100)
                        )
                    else:
                        aggregated = self.df.agg({col: agg_mode})[col]
                    properties[f"{col}_{agg_mode}"] = aggregated
        df = DataFrame([properties])
        traj_gdf = GeoDataFrame(df, crs=self.crs)
        return traj_gdf

    def to_mf_json(self, datetime_to_str=True, temporal_columns=None):
        """
        Converts a Trajectory to a dictionary compatible with the Moving
        Features JSON (MF-JSON) specification.

        Examples
        --------

        >>> traj.to_mf_json()

        Returns:
            dict: The MF-JSON representation of the GeoDataFrame as a dictionary.
        """
        tmp = self.to_point_gdf()
        t = tmp.index.name
        mf_json = gdf_to_mf_json(
            tmp.reset_index(),
            self.get_traj_id_col(),
            t,
            datetime_to_str=datetime_to_str,
            temporal_columns=temporal_columns,
        )
        return mf_json

    def get_start_location(self):
        """
        Return the trajectory's start location.

        Returns
        -------
        shapely Point
            Trajectory start location
        """
        return self.df.geometry.iloc[0]

    def get_end_location(self):
        """Return the trajectory's end location.

        Returns
        -------
        shapely Point
            Trajectory end location
        """
        return self.df.geometry.iloc[-1]

    def get_bbox(self):
        """
        Return the trajectory's bounding box.

        Returns
        -------
        tuple
            Bounding box values (minx, miny, maxx, maxy)
        """
        return tuple(self.df.total_bounds.tolist())  # (minx, miny, maxx, maxy)

    def get_start_time(self):
        """
        Return the trajectory's start time.

        Returns
        -------
        datetime.datetime
            Trajectory start time
        """
        return self.df.index.min().to_pydatetime()

    def get_end_time(self):
        """
        Return the trajectory's end time.

        Returns
        -------
        datetime.datetime
            Trajectory end time
        """
        return self.df.index.max().to_pydatetime()

    def get_duration(self):
        """
        Return the trajectory's duration from start to end.

        Returns
        -------
        datetime.timedelta
            Trajectory duration
        """
        return self.get_end_time() - self.get_start_time()

    def get_row_at(self, t, method="nearest"):
        """
        Return row of the trajectory's DataFrame at time t.

        Parameters
        ----------
        t : datetime.datetime
            Timestamp to extract a row for
        method : str
            Interpolation method (Pandas get_loc method)

        Returns
        -------
        Pandas series
            Row of the DataFrame at time t
        """
        try:
            return self.df.loc[t]
        except KeyError:
            index = self.df.index.sort_values().drop_duplicates()
            idx = index.get_indexer([t], method=method)[0]
            return self.df.iloc[idx]

    def interpolate_position_at(self, t):
        """
        Compute and return interpolated position at time t.

        Parameters
        ----------
        t : datetime.datetime
            Timestamp to interpolate at

        Returns
        -------
        shapely Point
            Interpolated position along the trajectory at time t
        """
        prev_row = self.get_row_at(t, "ffill")
        next_row = self.get_row_at(t, "bfill")
        t_diff = next_row.name - prev_row.name
        t_diff_at = t - prev_row.name
        line = LineString(
            [
                prev_row[self.get_geom_col()],
                next_row[self.get_geom_col()],
            ]
        )
        if t_diff == 0 or line.length == 0:
            return prev_row[self.get_geom_col()]
        interpolated_position = line.interpolate(t_diff_at / t_diff * line.length)
        return interpolated_position

    def get_position_at(self, t, method="interpolated"):
        """
        Compute and return position at time t.

        Parameters
        ----------
        t : datetime.datetime
            Timestamp to extract a row for
        method : str
            Interpolation method

        Returns
        -------
        shapely Point
            Position at time t

        Examples
        --------
        If the trajectory contains a position at the given timestamp, it is
        returned:

        >>> traj.get_position_at(datetime(2018, 1, 1, 12, 6))
        Point (6 0)

        If there is no trajectory position for the given timestamp, the default
        behaviour is to interpolate the location:

        >>> traj.get_position_at(datetime(2018, 1, 1, 12, 9))
        POINT (6 4.5)

        To get the trajectory position closest to the given timestamp, specify
        method='nearest':

        >>> traj.get_position_at(datetime(2018, 1, 1, 12, 9), method='nearest')
        POINT (6 6)
        """
        if t < self.get_start_time() or t > self.get_end_time():
            raise ValueError(
                f"Timestamp {t} outside the trajectory's time range "
                f"({self.get_start_time()} to {self.get_end_time()})"
            )
        if method not in ["nearest", "interpolated", "ffill", "bfill"]:
            raise ValueError(
                f"Invalid method {method}. Must be one of [nearest, interpolated, "
                "ffill, bfill]"
            )
        if method == "interpolated":
            return self.interpolate_position_at(t)
        else:
            row = self.get_row_at(t, method)
            try:
                return row[self.get_geom_col()][0]
            except TypeError:
                return row[self.get_geom_col()]

    def get_linestring_between(self, t1, t2, method="interpolated"):
        """
        Return LineString of segment between times t1 and t2.

        Parameters
        ----------
        t1 : datetime.datetime
            Start time for the segment
        t2 : datetime.datetime
            End time for the segment
        method : str
            Extraction method

        Returns
        -------
        shapely LineString
            Extracted trajectory segment
        """
        if method not in ["interpolated", "within"]:
            raise ValueError(
                f"Invalid split method {method}. Must be one of [interpolated, within]"
            )
        if method == "interpolated":
            st_range = STRange(
                self.get_position_at(t1), self.get_position_at(t2), t1, t2
            )
            temp_df = create_entry_and_exit_points(self, st_range)
            temp_df = temp_df[t1:t2]
            return point_gdf_to_linestring(temp_df, self.get_geom_col())
        else:
            try:
                return point_gdf_to_linestring(
                    self.get_segment_between(t1, t2).df, self.get_geom_col()
                )
            except RuntimeError:
                raise RuntimeError(f"Cannot generate linestring between {t1} and {t2}")

    def get_segment_between(self, t1, t2):
        """
        Return Trajectory segment between times t1 and t2.

        Parameters
        ----------
        t1 : datetime.datetime
            Start time for the segment
        t2 : datetime.datetime
            End time for the segment

        Returns
        -------
        Trajectory
            Extracted trajectory segment
        """
        segment = Trajectory(
            self.df[t1:t2],
            f"{self.id}_{t1}",
            parent=self,
            traj_id_col=self.get_traj_id_col(),
        )
        if not segment.is_valid():
            raise RuntimeError(
                f"Failed to extract valid trajectory segment between {t1} and {t2}"
            )
        return segment

    def _compute_distance(self, row, conversion):
        pt0 = row["prev_pt"]
        pt1 = row[self.get_geom_col()]
        if not isinstance(pt0, Point):
            return 0.0
        if not isinstance(pt1, Point):
            raise ValueError(f"Invalid trajectory! Got {pt1} instead of point!")
        if pt0 == pt1:
            return 0.0
        return measure_distance(pt0, pt1, self.is_latlon, conversion)

    def _add_prev_pt(self, force=True):
        """
        Create a shifted geometry column with previous positions.
        """
        if "prev_pt" not in self.df.columns or force:
            # TODO: decide on default enforcement behavior
            self.df = self.df.assign(prev_pt=self.df.geometry.shift())

    def get_length(self, units=UNITS()):
        """
        Return the length of the trajectory.

        Length is calculated using CRS units, except if the CRS is geographic
        (e.g. EPSG:4326 WGS84) then length is calculated in meters.

        If units have been declared:

        - For geographic projections, in declared units
        - For known CRS units, in declared units
        - For unknown CRS units, in declared units as if CRS is in meters

        Parameters
        ----------
        units : tuple(str)
            Units in which to calculate length values (default: CRS units)
            For more info, check the list of supported units at
            https://movingpandas.org/units

        Returns
        -------
        float
            Length of the trajectory
        """

        conversion = get_conversion(units, self.crs_units)
        return measure_length(self.df.geometry, self.is_latlon, conversion)

    def get_direction(self):
        """
        Return the direction of the trajectory.

        The direction is calculated between the trajectory's start and end
        location. Direction values are in degrees, starting North turning
        clockwise.

        Returns
        -------
        float
            Direction of the trajectory in degrees
        """
        pt0 = self.get_start_location()
        pt1 = self.get_end_location()
        if self.is_latlon:
            return calculate_initial_compass_bearing(pt0, pt1)
        else:
            return azimuth(pt0, pt1)

    def get_sampling_interval(self):
        """
        Return the sampling interval of the trajectory.

        The sampling interval is computed as the median time difference between
        consecutive rows in the trajectory's DataFrame.

        Returns
        -------
        datetime.timedelta
            Sampling interval
        """
        if hasattr(self, "timedelta_col_name"):
            if self.timedelta_col_name in self.df.columns:
                return self.df[self.timedelta_col_name].median()
        return self._get_df_with_timedelta()[TIMEDELTA_COL_NAME].median()

    def _compute_heading(self, row):
        pt0 = row["prev_pt"]
        pt1 = row[self.get_geom_col()]
        if not isinstance(pt0, Point):
            return 0.0
        if pt0 == pt1:
            return 0.0
        if self.is_latlon:
            return calculate_initial_compass_bearing(pt0, pt1)
        else:
            return azimuth(pt0, pt1)

    def _compute_angular_difference(self, row):
        degrees1 = row["prev_direction"]
        degrees2 = row["direction"]
        if degrees1 == degrees2:
            return 0.0
        else:
            return angular_difference(degrees1, degrees2)

    def _compute_speed(self, row, conversion):
        pt0 = row["prev_pt"]
        pt1 = row[self.get_geom_col()]
        if not isinstance(pt0, Point):
            return 0.0
        if not isinstance(pt1, Point):
            raise ValueError(f"Invalid trajectory! Got {pt1} instead of point!")
        if pt0 == pt1:
            return 0.0
        return get_speed2(pt0, pt1, row["delta_t"], self.is_latlon, conversion)

    def _connect_prev_pt_and_geometry(self, row):
        pt0 = row["prev_pt"]
        pt1 = row[self.get_geom_col()]
        if not isinstance(pt0, Point):
            return None
        if not isinstance(pt1, Point):
            raise ValueError(f"Invalid trajectory! Got {pt1} instead of point!")
        if pt0 == pt1:
            # to avoid intersection issues with zero length lines
            pt1 = translate(pt1, 0.00000001, 0.00000001)
        return LineString(list(pt0.coords) + list(pt1.coords))

    def add_traj_id(self, overwrite=False):
        """
        Add trajectory id column and values to the trajectory's DataFrame.

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing trajectory id values (default: False)
        """
        if TRAJ_ID_COL_NAME in self.df.columns and not overwrite:
            raise RuntimeError(
                f"Trajectory already contains a {TRAJ_ID_COL_NAME} column! "
                f"Use overwrite=True to overwrite exiting values."
            )
        self.df[TRAJ_ID_COL_NAME] = self.id
        return self

    def add_direction(self, overwrite=False, name=DIRECTION_COL_NAME):
        """
        Add direction column and values to the trajectory's DataFrame.

        The direction is calculated between consecutive locations.
        Direction values are in degrees, starting North turning clockwise.
        Values are [0, 360).

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing direction values (default: False)
        name : str
            Name of the direction column (default: "direction")
        """
        self.direction_col_name = name
        if self.direction_col_name in self.df.columns and not overwrite:
            raise RuntimeError(
                f"Trajectory already has a column named {self.direction_col_name}! "
                "Use overwrite=True to overwrite exiting values or update the "
                "name arg."
            )
        self._add_prev_pt()
        self.df[name] = self.df.apply(self._compute_heading, axis=1)
        # set the direction in the first row to the direction of the second row
        self.df.at[self.get_start_time(), name] = self.df.iloc[1][name]
        self.df.drop(columns=["prev_pt"], inplace=True)
        return self

    def add_angular_difference(
        self,
        overwrite=False,
        name=ANGULAR_DIFFERENCE_COL_NAME,
    ):
        """
        Add angular difference to the trajectory's DataFrame.

        Angular difference is calculated as the absolute smaller angle
        between direction for points along the trajectory.
        Values are [0, 180.0]
        """
        self.angular_difference_col_name = name
        if self.angular_difference_col_name in self.df.columns and not overwrite:
            raise RuntimeError(
                f"Trajectory already has a column named "
                f"{self.angular_difference_col_name}!"
                f"Use overwrite=True to overwrite exiting values or update the "
                f"name arg."
            )
        # Avoid computing direction again if already computed
        direction_col = self.get_direction_col()
        if direction_col in self.df.columns:
            direction_exists = True
            temp_df = self.df.copy()
        else:
            direction_exists = False
            self.add_direction(name=DIRECTION_COL_NAME)
            temp_df = self.df.copy()

        temp_df["prev_" + direction_col] = temp_df[direction_col].shift()
        self.df[name] = temp_df.apply(self._compute_angular_difference, axis=1)
        # set the first row to be 0
        self.df.at[self.get_start_time(), name] = 0.0
        if not direction_exists:
            self.df.drop(columns=[DIRECTION_COL_NAME], inplace=True)
        return self

    def add_distance(self, overwrite=False, name=DISTANCE_COL_NAME, units=None):
        """
        Add distance column and values to the trajectory's DataFrame.

        Distance values are computed between the current point and the previous:

        If no units have been declared:

        - For geographic projections (e.g. EPSG:4326 WGS84), in meters
        - For other projections, in CRS units

        If units have been declared:

        - For geographic projections, in declared units
        - For known CRS units, in declared units
        - For unknown CRS units, in declared units as if CRS is in meters

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing distance values (default: False)
        name : str
            Name of the distance column (default: "distance")
        units : str
            Units in which to calculate distance values (default: CRS units)
            For more info, check the list of supported units at
            https://movingpandas.org/units

        Examples
        ----------
        If no units are declared, the distance will be calculated
        in meters for geographic projections (e.g. EPSG:4326 WGS84)
        and in CRS units for all other projections

        >>>traj.add_distance()

        The default column name is "distance". If a column of this name
        already exists, a new column name can be specified

        >>>traj.add_distance(name="distance (CRS units)")

        If units are declared, the distance will be calculated in those units
        except if the CRS units are unknown, in which case the CRS units are
        assumed to be in meters

        >>>traj.add_distance(units="km")

        It is suggested to declare a name for the new column specifying units

        >>>traj.add_distance(name="distance (miles)", units="mi")
        >>>traj.add_distance(name="distance (US Survey Feet)", units="survey_ft")
        """
        self.distance_col_name = name
        if self.distance_col_name in self.df.columns and not overwrite:
            raise RuntimeError(
                f"Trajectory already has a column named {self.distance_col_name}! "
                "Use overwrite=True to overwrite exiting values or update the "
                "name arg."
            )
        conversion = get_conversion(units, self.crs_units)
        self.df = self._get_df_with_distance(conversion, name)
        return self

    def add_speed(self, overwrite=False, name=SPEED_COL_NAME, units=UNITS()):
        """
        Add speed column and values to the trajectory's DataFrame.

        Speed values are computed between the current point and the previous:

        If no units have been declared:

        - For geographic projections, in meters per second
        - For other projections, in CRS units per second

        If units have been declared:

        - For geographic projections, in declared units
        - For known CRS units, in declared units
        - For unknown CRS units, in declared units as if CRS distance
          units are meters

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing speed values (default: False)
        name : str
            Name of the speed column (default: "speed")
        units : tuple(str)
            Units in which to calculate speed

            distance : str
                Abbreviation for the distance unit
                (default: CRS units, or metres if geographic)
            time : str
                Abbreviation for the time unit (default: seconds)

            For more info, check the list of supported units at
            https://movingpandas.org/units

        Examples
        ----------
        If no units are declared, the speed will be calculated
        in meters per second for geographic projections (e.g. EPSG:4326 WGS84)
        and in CRS distance units per second for all other projections

        >>>traj.add_speed()

        The default column name is "speed". If a column of this name
        already exists, a new column name can be specified

        >>>traj.add_speed(name="speed (CRS units)")

        If units are declared, the speed will be calculated in those units
        except if the CRS distance units are unknown, in which case the
        CRS distance units are assumed to be meters

        >>>traj.add_speed(units=("km", "h"))

        It is suggested to declare a name for the new column specifying units

        >>>traj.add_speed(name="speed (mph)", units=("mi", "h"))
        >>>traj.add_speed(name="US Survey Feet/s", units=("survey_ft", "s"))
        """
        self.speed_col_name = name
        if self.speed_col_name in self.df.columns and not overwrite:
            raise RuntimeError(
                f"Trajectory already has a column named {self.speed_col_name}! "
                f"Use overwrite=True to overwrite exiting values or update the "
                f"name arg."
            )
        conversion = get_conversion(units, self.crs_units)
        self.df = self._get_df_with_speed(conversion, name)
        return self

    def add_acceleration(
        self, overwrite=False, name=ACCELERATION_COL_NAME, units=UNITS()
    ):
        """
        Add acceleration column and values to the trajectory's DataFrame.

        Acceleration values are computed between the current point and the previous:

        If no units have been declared:

        - For geographic projections, in meters per second squared
        - For other projections, in CRS units per second squared

        If units have been declared:

        - For geographic projections, using declared units
        - For known CRS units, using declared units
        - For unknown CRS units, using declared units as if CRS distance
          units are meters
        - If only distance units are declared, returns
          distance per second squared
        - If distance and one time unit declared, returns
          distance/time per second

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing speed values (default: False)
        name : str
            Name of the acceleration column (default: "acceleration")
        units : tuple(str)
            Units in which to calculate acceleration

            distance : str
                Abbreviation for the distance unit
                (default: CRS units, or metres if geographic)
            time : str
                Abbreviation for the time unit (default: seconds)
            time2 : str
                Abbreviation for the second time unit (default: seconds)

            For more info, check the list of supported units at
            https://movingpandas.org/units

        Examples
        ----------
        If no units are declared, the acceleration will be calculated
        in meters per second squared for geographic projections and
        in CRS distance units per second squared for all other projections

        >>>traj.add_acceleration()

        The default column name is "acceleration". If a column of this name
        already exists, a new column name can be specified

        >>>traj.add_acceleration(name="acceleration (CRS units/s2)")

        If units are declared, acceleration will be calculated in those units
        except if the CRS distance units are unknown, in which case the
        CRS distance units are assumed to be meters

        >>>traj.add_acceleration(units=("km", "h", "s"))

        It is suggested to declare a name for the new column specifying units

        >>>traj.add_acceleration(name="US Survey Feet/s2", units=("survey_ft", "s"))
        >>>traj.add_acceleration(name="mph/s", units=("mi", "h", "s"))
        """
        self.acceleration_col_name = name
        if self.acceleration_col_name in self.df.columns and not overwrite:
            raise RuntimeError(
                f"Trajectory already has a column named {self.acceleration_col_name}! "
                f"Use overwrite=True to overwrite exiting values or update the "
                f"name arg."
            )
        conversion = get_conversion(units, self.crs_units)
        self.df = self._get_df_with_acceleration(conversion, name)
        return self

    def add_timedelta(self, overwrite=False, name=TIMEDELTA_COL_NAME):
        """
        Add timedelta column and values to the trajectory's DataFrame.

        Timedelta is calculated as the time difference between the current
        and the previous row. Values are instances of datetime.timedelta.

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing timedelta values (default: False)
        name : str
            Name of the timedelta column (default: "timedelta")
        """
        self.timedelta_col_name = name
        if self.timedelta_col_name in self.df.columns and not overwrite:
            raise RuntimeError(
                f"Trajectory already has a column named {self.timedelta_col_name}! "
                f"Use overwrite=True to overwrite exiting values or update the "
                f"name arg."
            )
        self.df = self._get_df_with_timedelta(name)
        return self

    def _get_df_with_timedelta(self, name=TIMEDELTA_COL_NAME):
        temp_df = self.df.copy()
        times = Series(index=temp_df.index, data=temp_df.index)
        temp_df[name] = times.diff().values
        return temp_df

    def _get_df_with_distance(self, conversion, name=DISTANCE_COL_NAME):
        temp_df = self.df.copy()
        temp_df = temp_df.assign(prev_pt=temp_df.geometry.shift())
        try:
            temp_df[name] = temp_df.apply(
                self._compute_distance, conversion=conversion, axis=1
            )
        except ValueError as e:
            raise e
        # set the distance in the first row to zero
        temp_df.at[self.get_start_time(), name] = 0
        temp_df = temp_df.drop(columns=["prev_pt"])
        return temp_df

    def _get_df_with_speed(self, conversion, name=SPEED_COL_NAME):
        temp_df = self._get_df_with_timedelta(name="delta_t")
        temp_df = temp_df.assign(prev_pt=temp_df.geometry.shift())
        try:
            temp_df[name] = temp_df.apply(
                self._compute_speed, conversion=conversion, axis=1
            )
        except ValueError as e:
            raise e
        # set the speed in the first row to the speed of the second row
        temp_df.at[self.get_start_time(), name] = temp_df.iloc[1][name]
        temp_df = temp_df.drop(columns=["prev_pt", "delta_t"])
        return temp_df

    def _get_df_with_acceleration(self, conversion, name=ACCELERATION_COL_NAME):
        temp_df = self._get_df_with_speed(conversion, name="speed_temp")
        temp_df[name] = (
            temp_df["speed_temp"].diff()
            / temp_df.index.to_series().diff().dt.total_seconds()
            * conversion.time2
        )
        # set the acceleration in the first row to the acceleration of the
        # second row
        temp_df.at[self.get_start_time(), name] = temp_df.iloc[1][name]
        return temp_df.drop(columns=["speed_temp"])

    def intersects(self, polygon):
        """
        Return whether the trajectory intersects the given polygon.

        Parameters
        ----------
        polygon : shapely.geometry.Polygon
            Polygon to test for intersections

        Returns
        -------
        bool
        """
        return intersects(self, polygon)

    def distance(self, other, units=UNITS()):
        """
        Return the minimum distance to the other geometric object (based on shapely
        https://shapely.readthedocs.io/en/stable/manual.html#object.distance).

        If units have been declared:

        - For geographic projections, in declared units
        - For known CRS units, in declared units
        - For unknown CRS units, in declared units as if CRS is in meters

        Parameters
        ----------
        other : shapely.geometry or Trajectory
            Other geometric object or trajectory

        units : str
            Units in which to calculate distance values (default: CRS units)
            For more info, check the list of supported units at
            https://movingpandas.org/units

        Returns
        -------
        float
            Distance
        """
        if self.is_latlon:
            message = (
                f"Distance is computed using Euclidean geometry but "
                f"the trajectory coordinate system is {self.crs}."
            )
            warnings.warn(message, UserWarning)
        if type(other) == Trajectory:
            other = other.to_linestring()

        conversion = get_conversion(units, self.crs_units)
        return measure_distance_line(self.to_linestring(), other, conversion)

    def hausdorff_distance(self, other, units=UNITS()):
        """
        Return the Hausdorff distance to the other geometric object (based on shapely
        https://shapely.readthedocs.io/en/stable/manual.html#object.hausdorff_distance).
        The Hausdorff distance between two geometries is the furthest distance
        that a point on either geometry can be from the nearest point to it on
        the other geometry.

        If units have been declared:

        - For geographic projections, in declared units
        - For known CRS units, in declared units
        - For unknown CRS units, in declared units as if CRS is in meters

        Parameters
        ----------
        other : shapely.geometry or Trajectory
            Other geometric object or trajectory

        units : str
            Units in which to calculate distance values (default: CRS units)
            For more info, check the list of supported units at
            https://movingpandas.org/units

        Returns
        -------
        float
            Hausdorff distance
        """
        if self.is_latlon:
            message = (
                f"Hausdorff distance is computed using Euclidean geometry but "
                f"the trajectory coordinate system is {self.crs}."
            )
            warnings.warn(message, UserWarning)
        if type(other) == Trajectory:
            other = other.to_linestring()
        dist = self.to_linestring().hausdorff_distance(other)
        conversion = get_conversion(units, self.crs_units)
        return dist / conversion.distance

    def clip(self, polygon, point_based=False):
        """
        Return trajectory segments clipped by the given polygon.

        By default, the trajectory's line representation is clipped by the
        polygon. If pointbased=True, the trajectory's point representation is
        used instead, leading to shorter segments.

        Parameters
        ----------
        polygon : shapely Polygon
            Polygon to clip with
        point_based : bool
            Clipping method

        Returns
        -------
        TrajectoryCollection
            Clipped trajectory segments
        """
        from .trajectory_collection import TrajectoryCollection

        segments = clip(self, polygon, point_based)
        return TrajectoryCollection(segments)

    def intersection(self, feature, point_based=False):
        """
        Return the trajectory segments that intersects the given polygon feature.

        Feature attributes are appended to the trajectory's DataFrame.

        By default, the trajectory's line representation is clipped by the
        polygon. If pointbased=True, the trajectory's point representation is
        used instead, leading to shorter segments.

        Parameters
        ----------
        feature : shapely Feature
            Feature to intersect with
        point_based : bool
            Clipping method

        Returns
        -------
        TrajectoryCollection
            Segments intersecting with the feature
        """
        from .trajectory_collection import TrajectoryCollection

        segments = intersection(self, feature, point_based)
        return TrajectoryCollection(segments)

    def apply_offset_seconds(self, column, offset):
        """
        Shift column by the specified offset in seconds.

        Parameters
        ----------
        column : str
            Name of the column to shift
        offset : int
            Number of seconds to shift by, can be positive or negative
        """
        self.df[column] = self.df[column].shift(offset, freq="1s")

    def apply_offset_minutes(self, column, offset):
        """
        Shift column by the specified offset in minutes.

        Parameters
        ----------
        column : str
            Name of the column to shift
        offset : int
            Number of minutes to shift by, can be positive or negative
        """
        self.df[column] = self.df[column].shift(offset, freq="1min")

    def _to_line_df(self, columns=None):
        """
        Convert trajectory data GeoDataFrame of points to GeoDataFrame of lines
        that connect consecutive points.

        Returns
        -------
        line_df : GeoDataFrame
            GeoDataFrame of line segments
        """
        if columns is None:
            line_df = self.df.copy()
        else:
            line_df = self.df[columns].copy()
        line_df["prev_pt"] = line_df.geometry.shift()
        line_df["t"] = self.df.index
        line_df["prev_t"] = line_df["t"].shift()
        line_df["line"] = line_df.apply(self._connect_prev_pt_and_geometry, axis=1)
        line_df = line_df.set_geometry("line")[1:]
        return line_df

    def get_mcp(self):
        """Return the Minimum Convex Polygon of the trajectory data

        Returns
        -------
        mcp : Shapely object
            The polygon or line (in case of only two points)
            of the Minimum Convex Polygon
        """
        try:
            return self.df.geometry.union_all().convex_hull
        except AttributeError:
            return self.df.geometry.unary_union.convex_hull



# -*- coding: utf-8 -*-

from pytest import approx
from pandas.testing import assert_frame_equal
from shapely.geometry import Polygon
from datetime import datetime, timedelta
from movingpandas.tests.test_trajectory import Node, make_traj, CRS_METRIC, CRS_LATLON
from movingpandas.overlay import _get_potentially_intersecting_lines


class TestOverlay:
    def setup_method(self):
        self.nodes = [
            Node(0, 0, 1970, 1, 1, 0, 0, 0),
            Node(6, 0, 1970, 1, 1, 0, 0, 6),
            Node(10, 0, 1970, 1, 1, 0, 0, 10),
            Node(10, 10, 1970, 1, 1, 0, 0, 20),
            Node(0, 10, 1970, 1, 1, 0, 0, 30),
        ]
        self.default_traj_metric = make_traj(self.nodes[:3], CRS_METRIC)
        self.default_traj_latlon = make_traj(self.nodes[:3], CRS_LATLON)
        self.default_traj_metric_5 = make_traj(self.nodes, CRS_METRIC)

    def test_clip_one_intersections(self):
        polygon = Polygon([(5, -5), (7, -5), (7, 8), (5, 8), (5, -5)])
        traj = self.default_traj_metric_5
        intersections = traj.clip(polygon)
        assert len(intersections) == 1
        assert intersections.get_trajectory("1_0") == make_traj(
            [Node(5, 0, second=5), Node(6, 0, second=6), Node(7, 0, second=7)],
            id="1_0",
            parent=traj,
        )

    def test_clip_no_node_in_poly(self):
        polygon = Polygon([(1, -5), (2, -5), (2, 8), (1, 8), (1, -5)])
        traj = self.default_traj_metric_5
        intersections = traj.clip(polygon)
        assert len(intersections) == 1
        assert intersections.get_trajectory("1_0") == make_traj(
            [Node(1, 0, second=1), Node(2, 0, second=2)], id="1_0", parent=traj
        )

    def test_get_potentially_intersecting_lines(self):
        polygon = Polygon([(5, -5), (7, -5), (7, 8), (5, 8), (5, -5)])
        traj = self.default_traj_metric_5
        result = _get_potentially_intersecting_lines(traj, polygon)
        assert result.shape[0] == 2
        expected = make_traj(self.nodes[:3])._to_line_df()
        assert_frame_equal(expected, result)

    def test_clip_two_intersections_with_same_polygon(self):
        polygon = Polygon([(5, -5), (7, -5), (7, 12), (5, 12), (5, -5)])
        traj = self.default_traj_metric_5
        intersections = traj.clip(polygon)
        assert len(intersections) == 2
        assert intersections.get_trajectory("1_0") == make_traj(
            [Node(5, 0, second=5), Node(6, 0, second=6), Node(7, 0, second=7)],
            id="1_0",
            parent=traj,
        )
        assert intersections.get_trajectory("1_1") == make_traj(
            [Node(7, 10, second=23), Node(5, 10, second=25)], id="1_1", parent=traj
        )

    def test_clip_with_duplicate_traj_points_does_not_drop_any_points(self):
        polygon = Polygon([(5, -5), (7, -5), (7, 5), (5, 5), (5, -5)])
        traj = make_traj(
            [
                Node(),
                Node(6, 0, second=6),
                Node(6, 0, second=7),
                Node(10, 0, second=11),
                Node(10, 10, second=20),
                Node(0, 10, second=30),
            ]
        )
        intersections = traj.clip(polygon)
        assert len(intersections) == 1
        assert intersections.get_trajectory("1_0") == make_traj(
            [
                Node(5, 0, second=5),
                Node(6, 0, second=6),
                Node(6, 0, second=7),
                Node(7, 0, second=8),
            ],
            id="1_0",
            parent=traj,
        )

    def test_clip_pointbased(self):
        polygon = Polygon([(5.1, -5), (7.5, -5), (7.5, 12), (5.1, 12), (5.1, -5)])
        traj = make_traj(
            [
                Node(),
                Node(6, 0, minute=6),
                Node(6.5, 0, minute=6, second=30),
                Node(7, 0, minute=7),
                Node(10, 0, minute=10),
            ]
        )
        intersections = traj.clip(polygon, point_based=True)
        assert len(intersections) == 1
        assert intersections.get_trajectory("1_0") == make_traj(
            [
                Node(6, 0, minute=6),
                Node(6.5, 0, minute=6, second=30),
                Node(7, 0, minute=7),
            ],
            id="1_0",
            parent=traj,
        )

    def test_clip_pointbased_singlepoint_returns_empty(self):
        polygon = Polygon([(5.1, -5), (6.4, -5), (6.4, 12), (5.1, 12), (5.1, -5)])
        traj = make_traj(
            [
                Node(),
                Node(6, 0, minute=6),
                Node(6.5, 0, minute=6, second=30),
                Node(7, 0, minute=7),
                Node(10, 0, minute=10),
            ]
        )
        intersections = traj.clip(polygon, point_based=True)
        assert len(intersections) == 0

    def test_clip_interpolated_singlepoint(self):
        polygon = Polygon([(5.1, -5), (6.4, -5), (6.4, 12), (5.1, 12), (5.1, -5)])
        traj = make_traj(
            [
                Node(0, 0, minute=5),
                Node(6, 0, minute=6),
                Node(6.5, 0, minute=6, second=30),
            ]
        )
        intersections = traj.clip(polygon, point_based=False)
        assert len(intersections) == 1
        assert intersections.get_trajectory("1_0") == make_traj(
            [
                Node(5.1, 0, minute=5, second=51),
                Node(6, 0, minute=6),
                Node(6.4, 0, minute=6, second=24),
            ],
            id="1_0",
            parent=traj,
        )

    def test_clip_does_not_alter_df(self):
        polygon = Polygon([(5, -5), (7, -5), (7, 12), (5, 12), (5, -5)])
        traj = self.default_traj_metric_5.copy()
        traj.clip(polygon)
        assert_frame_equal(self.default_traj_metric_5.df, traj.df)

    def test_clip_with_one_intersection_reversed(self):
        polygon = Polygon([(5, -5), (7, -5), (7, 5), (5, 5), (5, -5)])
        traj = make_traj([Node(10, 0), Node(6, 0, minute=10), Node(0, 0, minute=20)])
        intersections = traj.clip(polygon)
        assert len(intersections) == 1
        assert intersections.get_trajectory("1_0") == make_traj(
            [
                Node(7, 0, minute=7, second=30),
                Node(6, 0, minute=10),
                Node(5, 0, minute=11, second=40),
            ],
            id="1_0",
            parent=traj,
        )

    def test_clip_with_milliseconds(self):
        polygon = Polygon([(5, -5), (7, -5), (8, 5), (5, 5), (5, -5)])
        traj = make_traj(
            [
                Node(0, 10, hour=12),
                Node(10, 10, hour=12, minute=10),
                Node(10, 0, hour=12, minute=15),
                Node(6, 0, hour=12, minute=30),
                Node(0, 0, hour=13),
            ]
        )
        intersection = traj.clip(polygon).get_trajectory("1_0")
        assert intersection.to_linestring().wkt == "LINESTRING (7.5 0, 6 0, 5 0)"
        assert intersection.get_start_time() - datetime(
            1970, 1, 1, 12, 24, 22, 500000
        ) < timedelta(milliseconds=1)
        assert intersection.get_end_time() == datetime(1970, 1, 1, 12, 35, 0)

    def test_clip_with_numerical_time_issues(self):
        xmin, xmax, ymin, ymax = (
            116.36850352835575,
            116.37029459899574,
            39.904675309969896,
            39.90772814977718,
        )
        polygon = Polygon(
            [(xmin, ymin), (xmin, ymax), (xmax, ymax), (xmax, ymin), (xmin, ymin)]
        )
        traj = make_traj(
            [
                Node(116.36855, 39.904926, 2009, 3, 10, 11, 3, 35),
                Node(116.368612, 39.904877, 2009, 3, 10, 11, 3, 37),
                Node(116.368644, 39.90484, 2009, 3, 10, 11, 3, 39),
            ]
        )
        result = traj.clip(polygon).get_trajectory("1_0").to_linestring().wkt
        assert (
            result == "LINESTRING (116.36855 39.904926, "
            "116.368612 39.904877, "
            "116.368644 39.90484)"
        )

    def test_clip_with_no_intersection(self):
        polygon = Polygon([(105, -5), (107, -5), (107, 12), (105, 12), (105, -5)])
        intersections = self.default_traj_metric.clip(polygon)
        assert len(intersections) == 0

    def test_intersection_with_feature(self):
        feature = {
            "geometry": {
                "type": "Polygon",
                "coordinates": [[(5, -5), (7, -5), (8, 5), (5, 5), (5, -5)]],
            },
            "properties": {"id": 1, "name": "foo"},
        }
        intersection = self.default_traj_metric_5.intersection(feature).get_trajectory(
            "1_0"
        )
        assert len(intersection.df.columns) == len(
            ["geometry", "value", "traj_id", "intersecting_id", "intersecting_name"]
        )
        assert intersection.df.iloc[0]["intersecting_id"] == 1
        assert intersection.df.iloc[0]["intersecting_name"] == "foo"

    def test_clip_with_empty_spatial_intersection_linestrings(self):
        polygon = Polygon(
            [
                (-92.98830, 35.38896),
                (-92.35109, 33.96139),
                (-93.27393, 33.61437),
                (-93.99904, 34.88572),
                (-92.98830, 35.38896),
            ]
        )
        traj = make_traj(
            [
                Node(-92.00000, 35.35000, 2020, 8, 5, 21, 33),
                Node(-92.10000, 35.30000, 2020, 8, 5, 21, 34),
                Node(-92.48333, 35.05000, 2020, 8, 5, 21, 38),
                Node(-92.68333, 34.95000, 2020, 8, 5, 21, 40),
                Node(-92.80000, 34.86667, 2020, 8, 5, 21, 41),
                Node(-93.15000, 34.66667, 2020, 8, 5, 21, 44),
                Node(-93.31667, 34.55000, 2020, 8, 5, 21, 45),
                Node(-93.66667, 34.35000, 2020, 8, 5, 21, 48),
                Node(-93.88333, 34.20000, 2020, 8, 5, 21, 50),
                Node(-94.00000, 34.16904, 2020, 8, 5, 21, 51),
            ]
        )
        result = traj.clip(polygon)
        assert len(result) == 1
        coords = result.get_trajectory("1_0").to_linestring().coords
        expected = [
            [-92.76600766934008, 34.89094857130274],
            [-92.8, 34.86667],
            [-93.15000000000001, 34.66667],
            [-93.31667, 34.55],
            [-93.66667, 34.35],
            [-93.68590048921358, 34.33668617473444],
        ]
        assert len(coords) == len(expected)
        for o, e in zip(coords, expected):
            assert o == approx(e, 0.1)



# -*- coding: utf-8 -*-

from collections import Counter

from pandas import DataFrame
from geopandas import GeoDataFrame
from shapely.geometry import LineString
from shapely.ops import nearest_points

from movingpandas.point_clusterer import PointClusterer
from .geometry_utils import (
    azimuth,
    angular_difference,
    measure_distance,
)


class TrajectoryCollectionAggregator:
    def __init__(
        self,
        traj_collection,
        max_distance,
        min_distance,
        min_stop_duration,
        min_angle=45,
    ):
        """
        Aggregates trajectories by extracting significant points,
        clustering those points, and extracting flows between clusters.

        Parameters
        ----------
        traj_collection : TrajectoryCollection
            TrajectoryCollection to be aggregated
        max_distance : float
            Maximum distance between significant points (distance is
            calculated in CRS units, except if the CRS is geographic, e.g.
            EPSG:4326 WGS84, then distance is calculated in meters)
        min_distance : float
            Minimum distance between significant points
        min_stop_duration : datetime.timedelta
            Minimum duration required for stop detection
        min_angle : float
            Minimum angle for significant point extraction

        References
        ----------
        * Andrienko, N., & Andrienko, G. (2011). Spatial generalization and
          aggregation of massive movement data. IEEE Transactions on
          visualization and computer graphics, 17(2), 205-219.
        """
        self.traj_collection = traj_collection
        if self.traj_collection.trajectories:
            self._crs = traj_collection.trajectories[0].crs
        else:
            self._crs = None
        self.max_distance = max_distance
        self.min_distance = min_distance
        self.min_stop_duration = min_stop_duration
        self.min_angle = min_angle
        self.is_latlon = self.traj_collection.trajectories[0].is_latlon
        # print('Extracting significant points ...')
        self.significant_points = self._extract_significant_points()
        # print('Clustering significant points ...')
        self.clusters = PointClusterer(
            self.significant_points, self.max_distance, self.is_latlon
        ).get_clusters()
        # print('  No. clusters: {}'.format(len(self.clusters)))
        # print('Computing flows ...')
        self.flows = self._compute_flows_between_clusters()
        # print('Flows ready!')

    def get_significant_points_gdf(self):
        """
        Return the extracted significant points

        Returns
        -------
        GeoDataFrame
            Significant points
        """
        if not self.significant_points:
            self.significant_points = self._extract_significant_points()
        df = DataFrame(self.significant_points, columns=["geometry"])
        return GeoDataFrame(df, crs=self._crs)

    def get_clusters_gdf(self):
        """
        Return the extracted cluster centroids

        Returns
        -------
        GeoDataFrame
            Cluster centroids, incl. the number of clustered significant
            points (n).
        """
        if not self.clusters:
            self.clusters = PointClusterer(
                self.significant_points, self.max_distance, self.is_latlon
            ).get_clusters()
        df = DataFrame(
            [cluster.centroid for cluster in self.clusters],
            columns=["geometry"],
        )
        df["n"] = [len(cluster.points) for cluster in self.clusters]
        return GeoDataFrame(df, crs=self._crs)

    def get_flows_gdf(self):
        """
        Return the extracted flows

        Returns
        -------
        GeoDataFrame
            Flow lines, incl. the number of trajectories summarized in the
            flow (weight) and the number of unique trajectory objects summarized
            in the flow (obj_weight).
        """
        if not self.flows:
            self.flows = self._compute_flows_between_clusters()
        return GeoDataFrame(self.flows, crs=self._crs)

    def _extract_significant_points(self):
        sig_points = []
        for traj in self.traj_collection:
            a = _PtsExtractor(
                traj,
                self.max_distance,
                self.min_distance,
                self.min_stop_duration,
                self.min_angle,
            )
            sig_points = sig_points + a.find_significant_points()
        return sig_points

    def _compute_flows_between_clusters(self):
        sg = _SequenceGenerator(self.get_clusters_gdf(), self.traj_collection)
        return sg.create_flow_lines()


class _PtsExtractor:
    def __init__(
        self, traj, max_distance, min_distance, min_stop_duration, min_angle=45
    ):
        self.traj = traj
        self.traj_geom = traj.df[traj.get_geom_col()]
        self.n = self.traj.df.geometry.count()
        self.max_distance = max_distance
        self.min_distance = min_distance
        self.min_stop_duration = min_stop_duration
        self.min_angle = min_angle
        self.significant_points = [
            self.traj.get_start_location(),
            self.traj.get_end_location(),
        ]

    def find_significant_points(self):
        i = 0
        j = 1
        while j < self.n - 1:
            if self.is_significant_distance(i, j):
                self.add_point(j)
                i = j
                j = i + 1
                continue
            k, has_points = self.locate_points_beyond_min_distance(j)
            if has_points:
                # print(f"has points (i={i} | j={j} | k={k})")
                if k > j + 1:
                    if self.is_significant_stop(j, k):
                        self.add_point(j)
                        i = j
                        j = k
                        continue
                    # compute the average spatial position to represent
                    # the similar points
                    else:
                        m = j + (k - 1 - j) / 2
                        j = int(m)
                if self.is_significant_turn(i, j, k):
                    self.add_point(j)
                    i = j
                    j = k
                else:
                    j += 1
            else:
                return self.significant_points
        return self.significant_points

    def is_significant_turn(self, i, j, k):
        angle = self.compute_angle_between_vectors(i, j, k)
        return angle >= self.min_angle and angle <= (360 - self.min_angle)

    def is_significant_stop(self, j, k):
        delta_t = self.traj.df.iloc[k - 1].name - self.traj.df.iloc[j].name
        return delta_t >= self.min_stop_duration

    def add_point(self, j):
        pt = self.get_pt(j)
        self.append_point(pt)

    def append_point(self, pt):
        if pt != self.traj.get_start_location():
            self.significant_points.append(pt)

    def compute_angle_between_vectors(self, i, j, k):
        p_i = self.get_pt(i)
        p_j = self.get_pt(j)
        p_k = self.get_pt(k)
        azimuth_ij = azimuth(p_i, p_j)
        azimuth_jk = azimuth(p_j, p_k)
        return angular_difference(azimuth_ij, azimuth_jk)

    def locate_points_beyond_min_distance(self, j):
        for k in range(j + 1, self.n):
            if self.distance_greater_than(j, k, self.min_distance):
                return k, True
        return k, False

    def distance_greater_than(self, loc1, loc2, dist):
        pt1 = self.get_pt(loc1)
        pt2 = self.get_pt(loc2)
        d = measure_distance(pt1, pt2, self.traj.is_latlon)
        return d >= dist

    def get_pt(self, the_loc):
        return self.traj_geom.iloc[the_loc]

    def is_significant_distance(self, i, j):
        if self.distance_greater_than(i, j, self.max_distance):
            self.append_point(self.get_pt(i))
            return True
        return False


class _SequenceGenerator:
    def __init__(self, cells, traj_collection):
        self.cells = cells
        try:
            self.cells_union = cells.geometry.union_all()
        except AttributeError:
            self.cells_union = cells.geometry.unary_union

        self.id_to_centroid = {i: [f, [0, 0, 0, 0, 0]] for i, f in cells.iterrows()}
        self.sequences = Counter()
        self.sequences_obj_log = {}
        for traj in traj_collection:
            self.evaluate_trajectory(traj)

    def evaluate_trajectory(self, trajectory):
        this_sequence = []
        prev_cell_id = None
        geom_name = trajectory.get_geom_col()
        for t, geom in trajectory.df[geom_name].items():
            nearest_id = self.get_nearest(geom)
            nearest_cell = self.id_to_centroid[nearest_id][0]
            nearest_cell_id = nearest_cell.name
            if this_sequence:
                prev_cell_id = this_sequence[-1]
                if nearest_cell_id != prev_cell_id:
                    self.sequences[(prev_cell_id, nearest_cell_id)] += 1
                    if (prev_cell_id, nearest_cell_id) in self.sequences_obj_log.keys():
                        self.sequences_obj_log[(prev_cell_id, nearest_cell_id)].update(
                            [trajectory.obj_id]
                        )
                    else:
                        self.sequences_obj_log[(prev_cell_id, nearest_cell_id)] = set(
                            [trajectory.obj_id]
                        )
            if nearest_cell_id != prev_cell_id:
                # we have changed to a new cell --> up the counter
                h = t.hour
                self.id_to_centroid[nearest_id][1][0] += 1
                self.id_to_centroid[nearest_id][1][h // 6 + 1] += 1
                this_sequence.append(nearest_cell_id)

    def create_flow_lines(self):
        lines = []
        for key, value in self.sequences.items():
            p1 = self.id_to_centroid[key[0]][0].geometry
            p2 = self.id_to_centroid[key[1]][0].geometry
            obj_value = len(self.sequences_obj_log[key])
            lines.append(
                {
                    "geometry": LineString([p1, p2]),
                    "weight": value,
                    "obj_weight": obj_value,
                }
            )
        return lines

    def get_nearest(self, pt):
        nearest = self.cells.geometry.geom_equals(
            nearest_points(pt, self.cells_union)[1]
        )
        return self.cells[nearest].iloc[0].name



# -*- coding: utf-8 -*-

import itertools as it
import pandas as pd
from shapely.geometry import Point, LineString, shape
from shapely.affinity import translate
from datetime import datetime, timedelta

from .spatiotemporal_utils import TRange, STRange


def _get_spatiotemporal_ref(row):
    """
    Returns the SpatioTemporalRange for the input row's spatial_intersection
    LineString by interpolating timestamps.
    """
    if row["spatial_intersection"].is_empty:
        return None
    if type(row["spatial_intersection"]) == LineString:
        pt0 = Point(row["spatial_intersection"].coords[0])
        ptn = Point(row["spatial_intersection"].coords[-1])
        t = row["prev_t"]
        t_delta = row["t"] - t
        length = row["line"].length
        t0 = t + (t_delta * row["line"].project(pt0) / length)
        tn = t + (t_delta * row["line"].project(ptn) / length)
        # to avoid numerical issues with microseconds beyond six digits,
        # we reconstruct the timestamps
        t0 = datetime(
            t0.year, t0.month, t0.day, t0.hour, t0.minute, t0.second, t0.microsecond
        )
        tn = datetime(
            tn.year, tn.month, tn.day, tn.hour, tn.minute, tn.second, tn.microsecond
        )
        # to avoid intersection issues with zero length lines
        if ptn == translate(pt0, 0.00000001, 0.00000001):
            t0 = row["prev_t"]
            tn = row["t"]
        # to avoid numerical issues with timestamps
        if is_equal(tn, row["t"]):
            tn = row["t"]
        if is_equal(t0, row["prev_t"]):
            t0 = row["prev_t"]
        return STRange(pt0, ptn, t0, tn)
    else:
        return None


def _dissolve_ranges(ranges):
    """
    SpatioTemporalRanges that touch (i.e. the end of one equals the start of
    another) are dissolved (aka. merged).
    """
    if len(ranges) == 0:
        raise ValueError("Nothing to dissolve (received empty ranges)!")
    dissolved_ranges = []
    new_range = None
    for r in ranges:
        if r is None:
            continue  # raise ValueError('Received range that is None!')
        if new_range is None:
            new_range = STRange(r.pt_0, r.pt_n, r.t_0, r.t_n)
        elif new_range.t_n == r.t_0 or (
            r.t_0 > new_range.t_n and is_equal(r.t_0, new_range.t_n)
        ):
            new_range.t_n = r.t_n
            new_range.pt_n = r.pt_n
        else:
            dissolved_ranges.append(new_range)
            new_range = STRange(r.pt_0, r.pt_n, r.t_0, r.t_n)
    dissolved_ranges.append(new_range)
    return dissolved_ranges


def is_equal(t1, t2):
    """
    Similar timestamps are considered equal to avoid numerical issues.
    """
    if type(t2) == pd.Timestamp:
        td = abs(t1 - t2.tz_localize(t1.tzinfo).to_pydatetime())
    else:
        td = abs(t1 - t2)
    return td < timedelta(milliseconds=10)


def intersects(traj, polygon):
    try:
        line = traj.to_linestring()
    except:  # noqa: E722
        return False
    return line.intersects(polygon)


def create_entry_and_exit_points(traj, range):
    """
    Returns a dataframe with inserted entry and exit points according to the
    provided SpatioTemporalRange.
    """
    if type(range) != STRange:
        raise TypeError("Input range has to be a SpatioTemporalRange!")

    crs = traj.df.crs
    temp_df = traj.df.copy()

    index = traj.df.index
    # Create row at entry point with attributes from previous row = pad
    row0 = traj.df.iloc[index.get_indexer([range.t_0], method="pad")[0]].copy()
    row0["geometry"] = range.pt_0
    # Create row at exit point
    rown = traj.df.iloc[index.get_indexer([range.t_n], method="pad")[0]].copy()
    rown["geometry"] = range.pt_n
    # Insert rows
    try:
        temp_df.loc[range.t_0] = row0
    except ValueError as err:
        if str(err) == "cannot set a single element with an array":
            # fix for https://github.com/anitagraser/movingpandas/issues/118
            # (not sure what causes this problem)
            pass
        else:
            raise err
    try:
        temp_df.loc[range.t_n] = rown
    except ValueError as err:
        if str(err) == "cannot set a single element with an array":
            pass
        else:
            raise err

    # ensure CRS is set, fix for https://github.com/anitagraser/movingpandas/issues/291
    temp_df = temp_df.set_crs(crs, allow_override=True)

    return temp_df.sort_index()


def _get_segments_for_ranges(traj, ranges):
    counter = it.count()
    segments = []  # list of trajectories
    for the_range in ranges:
        temp_traj = traj.copy()
        if type(the_range) == STRange:
            temp_traj.df = create_entry_and_exit_points(traj, the_range)
        try:
            segment = temp_traj.get_segment_between(the_range.t_0, the_range.t_n)
        except ValueError:
            continue
        segment.id = f"{traj.id}_{next(counter)}"
        segment.parent = traj
        segments.append(segment)
    return segments


def _determine_time_ranges_pointbased(traj, polygon):
    df = traj.df
    df["t"] = df.index
    df["intersects"] = df.intersects(polygon)
    df["segment"] = (df["intersects"].shift(1) != df["intersects"]).astype(int).cumsum()
    df = df.groupby("segment", as_index=False).agg(
        {"t": ["min", "max"], "intersects": ["min"]}
    )
    df.columns = df.columns.map("_".join)

    ranges = []
    for _, row in df[df["intersects_min"]].iterrows():
        ranges.append(TRange(row["t_min"], row["t_max"]))
    return ranges


def _get_potentially_intersecting_lines(traj, polygon):
    """
    Uses a spatial index to determine which parts of the trajectory may be
    intersecting with the polygon

    Returns
    -------
    possible_matches : GeoDataFrame
        GeoDataFrame of potentially intersecting lines
    """
    line_df = traj._to_line_df()
    spatial_index = line_df.sindex
    if spatial_index:
        possible_matches_index = list(spatial_index.intersection(polygon.bounds))
        possible_matches = line_df.iloc[possible_matches_index].sort_index()
    else:
        possible_matches = line_df
    return possible_matches


def _determine_time_ranges_linebased(traj, polygon):
    """
    Returns list of SpatioTemporalRanges that describe trajectory intersections
    with the provided polygon.
    """
    # Note: If the trajectory contains consecutive rows without location change
    #       these will result in zero length lines that return an empty
    #       intersection.
    possible_matches = _get_potentially_intersecting_lines(traj, polygon)
    possible_matches["spatial_intersection"] = possible_matches.intersection(polygon)
    possible_matches["spatiotemporal_intersection"] = possible_matches.apply(
        _get_spatiotemporal_ref, axis=1
    )
    ranges = possible_matches["spatiotemporal_intersection"]
    return _dissolve_ranges(ranges)


def clip(traj, polygon, pointbased=False):
    """
    Returns a list of trajectory segments clipped by the given feature.
    """
    if not intersects(traj, polygon):
        return []
    if pointbased:
        ranges = _determine_time_ranges_pointbased(traj, polygon)
    else:
        ranges = _determine_time_ranges_linebased(traj, polygon)
    return _get_segments_for_ranges(traj, ranges)


def _get_geometry_and_properties_from_feature(feature):
    """
    Provides convenience access to geometry and properties of a Shapely feature.
    """
    if type(feature) != dict:
        raise TypeError("Trajectories can only be intersected with a Shapely feature!")
    try:
        geometry = shape(feature["geometry"])
        properties = feature["properties"]
    except:  # noqa: E722
        raise TypeError("Trajectories can only be intersected with a Shapely feature!")
    return geometry, properties


def intersection(traj, feature, pointbased=False):
    """
    Returns a list of trajectory segments that intersect the given feature.
    Resulting trajectories include the intersecting feature's attributes.
    """
    geometry, properties = _get_geometry_and_properties_from_feature(feature)
    clipped = clip(traj, geometry, pointbased)
    segments = []
    for clipping in clipped:
        for key, value in properties.items():
            clipping.df["intersecting_" + key] = value
        segments.append(clipping)
    return segments



# -*- coding: utf-8 -*-

import pytest
from math import sqrt
from shapely.geometry import MultiPoint, Point
from movingpandas.geometry_utils import (
    azimuth,
    calculate_initial_compass_bearing,
    angular_difference,
    mrr_diagonal,
    measure_distance_geodesic,
    measure_distance_euclidean,
    measure_distance_spherical,
)


class TestGeometryUtils:
    def test_compass_bearing_east(self):
        assert calculate_initial_compass_bearing(Point(0, 0), Point(10, 0)) == 90

    def test_compass_bearing_west(self):
        assert calculate_initial_compass_bearing(Point(0, 0), Point(-10, 0)) == 270

    def test_compass_bearing_north(self):
        assert calculate_initial_compass_bearing(Point(0, 0), Point(0, 10)) == 0

    def test_compass_bearing_south(self):
        assert calculate_initial_compass_bearing(Point(0, 0), Point(0, -10)) == 180

    def test_azimuth_east(self):
        assert azimuth(Point(0, 0), Point(1, 0)) == 90
        assert azimuth(Point(0, 0), Point(100, 0)) == 90

    def test_azimuth_west(self):
        assert azimuth(Point(0, 0), Point(-10, 0)) == 270

    def test_azimuth_north(self):
        assert azimuth(Point(0, 0), Point(0, 1)) == 0

    def test_azimuth_south(self):
        assert azimuth(Point(0, 0), Point(0, -1)) == 180

    def test_azimuth_northeast(self):
        assert azimuth(Point(0, 0), Point(1, 1)) == 45

    def test_azimuth_southeast(self):
        assert azimuth(Point(0, 0), Point(1, -1)) == 135

    def test_azimuth_southwest(self):
        assert azimuth(Point(0, 0), Point(-1, -1)) == 225

    def test_azimuth_northwest(self):
        assert azimuth(Point(100, 100), Point(99, 101)) == 315

    def test_anglular_difference_tohigher(self):
        assert angular_difference(1, 5) == 4

    def test_anglular_difference_tolower(self):
        assert angular_difference(355, 5) == 10

    def test_anglular_difference_halfcicle(self):
        assert angular_difference(180, 0) == 180

    def test_anglular_difference_same(self):
        assert angular_difference(45, 45) == 0

    def test_anglular_difference_onenegative(self):
        assert angular_difference(-45, 45) == 90

    def test_anglular_difference_twonegative(self):
        assert angular_difference(-200, -160) == 40

    def test_mrr_diagonal(self):
        assert mrr_diagonal(
            MultiPoint([Point(0, 0), Point(0, 2), Point(2, 0), Point(2, 2)])
        ) == sqrt(8)

    def test_mrr_diagonal_one_point(self):
        assert mrr_diagonal(Point(2, 3)) == 0

    def test_euclidean_distance(self):
        assert (measure_distance_euclidean(Point(0, 0), Point(0, 1))) == 1

    def test_spherical_distance(self):
        assert measure_distance_spherical(
            Point(-74.00597, 40.71427), Point(-118.24368, 34.05223)
        ) == pytest.approx(3935735)

    def test_geodesic_distance(self):
        # Distance between NYC, NY USA and Los Angeles, CA USA is
        # 3944411.0951634306 meters
        assert measure_distance_geodesic(
            Point(-74.00597, 40.71427), Point(-118.24368, 34.05223)
        ) == pytest.approx(3944411)

    def test_measure_distance_euclidean_throws_type_error(self):
        with pytest.raises(TypeError):
            measure_distance_euclidean((0, 0), (0, 1))

    def test_measure_distance_spherical_throws_type_error(self):
        with pytest.raises(TypeError):
            measure_distance_spherical((0, 0), (0, 1))

    def test_measure_distance_geodesic_throws_type_error(self):
        with pytest.raises(TypeError):
            measure_distance_geodesic((0, 0), (0, 1))

    def test_calculate_initial_compass_bearing_throws_type_error(self):
        with pytest.raises(TypeError):
            calculate_initial_compass_bearing((0, 0), (0, 1))

    def test_azimuth_throws_type_error(self):
        with pytest.raises(TypeError):
            azimuth((0, 0), (0, 1))

    def test_mrr_diagonal_linestring(self):
        # Distance between NYC, NY USA and Los Angeles, CA USA is
        # 3944411.0951634306 meters
        assert mrr_diagonal(
            MultiPoint(
                [
                    Point(-74.00597, 40.71427),
                    Point(-74.00597, 40.71427),
                    Point(-118.24368, 34.05223),
                ]
            ),
            spherical=True,
        ) == pytest.approx(3944411)



# -*- coding: utf-8 -*-

from copy import copy
from datetime import datetime as datetime, timedelta as timedelta
from math import sqrt

import pandas as pd
import pytest
from pyproj import CRS
from geopandas import GeoDataFrame
from pandas import Timestamp
from pandas.testing import assert_frame_equal
from shapely.geometry import LineString, Point, Polygon, MultiPolygon

from movingpandas.trajectory import (
    TRAJ_ID_COL_NAME,
    SPEED_COL_NAME,
    DIRECTION_COL_NAME,
    DISTANCE_COL_NAME,
    ACCELERATION_COL_NAME,
    ANGULAR_DIFFERENCE_COL_NAME,
    TIMEDELTA_COL_NAME,
)
from movingpandas.trajectory_collection import TrajectoryCollection
from . import requires_holoviews, requires_folium, has_geopandas1

CRS_METRIC = CRS.from_user_input(31256)
CRS_LATLON = CRS.from_user_input(4326)


class TestTrajectoryCollection:
    def setup_method(self):
        df = pd.DataFrame(
            [
                [1, "A", Point(0, 0), datetime(2018, 1, 1, 12, 0, 0), 9, "a"],
                [1, "A", Point(6, 0), datetime(2018, 1, 1, 12, 6, 0), 5, "b"],
                [1, "A", Point(6, 6), datetime(2018, 1, 1, 14, 10, 0), 2, "c"],
                [1, "A", Point(9, 9), datetime(2018, 1, 1, 14, 15, 0), 4, "d"],
                [2, "A", Point(10, 10), datetime(2018, 1, 1, 12, 0, 0), 10, "e"],
                [2, "A", Point(16, 10), datetime(2018, 1, 1, 12, 6, 0), 6, "f"],
                [2, "A", Point(16, 16), datetime(2018, 1, 2, 13, 10, 0), 7, "g"],
                [2, "A", Point(190, 10), datetime(2018, 1, 2, 13, 15, 0), 3, "h"],
            ],
            columns=["id", "obj", "geometry", "t", "val", "val2"],
        ).set_index("t")
        self.geo_df = GeoDataFrame(df, crs=CRS_METRIC)
        self.collection = TrajectoryCollection(self.geo_df, "id", obj_id_col="obj")
        self.geo_df_latlon = GeoDataFrame(df).set_crs(
            crs=CRS_LATLON, allow_override=True
        )
        self.collection_latlon = TrajectoryCollection(
            self.geo_df_latlon, "id", obj_id_col="obj"
        )

    def test_number_of_trajectories(self):
        assert len(self.collection) == 2

    def test_number_of_trajectories_min_length(self):
        collection = TrajectoryCollection(
            self.geo_df, "id", obj_id_col="obj", min_length=100
        )
        assert len(collection) == 1

    def test_number_of_trajectories_min_length_never_reached(self):
        collection = TrajectoryCollection(
            self.geo_df, "id", obj_id_col="obj", min_length=1000
        )
        assert len(collection) == 0

    def test_number_of_trajectories_min_duration(self):
        collection = TrajectoryCollection(
            self.geo_df, "id", obj_id_col="obj", min_duration=timedelta(days=1)
        )
        assert len(collection) == 1

    def test_number_of_trajectories_min_duration_from_list(self):
        collection = TrajectoryCollection(
            self.collection.trajectories, min_duration=timedelta(days=1)
        )
        assert len(collection) == 1

    def test_number_of_trajectories_min_duration_never_reached(self):
        collection = TrajectoryCollection(
            self.geo_df, "id", obj_id_col="obj", min_duration=timedelta(weeks=1)
        )
        assert len(collection) == 0

    def test_get_trajectory(self):
        assert self.collection.get_trajectory(1).id == 1
        assert self.collection.get_trajectory(1).obj_id == "A"
        assert self.collection.get_trajectory(2).id == 2
        assert self.collection.get_trajectory(3) is None

    def test_get_trajectories(self):
        assert len(self.collection.get_trajectories("A")) == 2
        assert len(self.collection.get_trajectories("A").to_point_gdf()) == 8
        assert len(self.collection.get_trajectories("B")) == 0

    def test_get_locations_at(self):
        locs = self.collection.get_locations_at(datetime(2018, 1, 1, 12, 6, 0))
        assert len(locs) == 2
        assert locs.iloc[0].geometry in [Point(6, 0), Point(16, 10)]
        assert locs.iloc[0].val in [5, 6]
        assert locs.iloc[1].geometry in [Point(6, 0), Point(16, 10)]
        assert locs.iloc[0].geometry != locs.iloc[1].geometry

    def test_get_locations_at_nearest(self):
        locs = self.collection.get_locations_at(datetime(2018, 1, 1, 12, 6, 1))
        assert len(locs) == 2
        assert locs.iloc[0].val in [5, 6]

    def test_get_locations_at_out_of_time_range(self):
        locs = self.collection.get_locations_at(datetime(2017, 1, 1, 12, 6, 1))
        assert len(locs) == 0

    def test_get_start_locations(self):
        locs = self.collection.get_start_locations()
        assert len(locs) == 2
        assert locs.iloc[0].geometry in [Point(0, 0), Point(10, 10)]
        assert locs.iloc[0].id in [1, 2]
        assert locs.iloc[0].obj == "A"
        assert locs.iloc[0].val in [9, 10]
        assert locs.iloc[0].val2 in ["a", "e"]
        assert locs.iloc[1].geometry in [Point(0, 0), Point(10, 10)]
        assert locs.iloc[0].geometry != locs.iloc[1].geometry
        assert isinstance(locs, GeoDataFrame)
        assert locs.crs == CRS_METRIC

    def test_timestamp_column_present_in_start_locations(self):
        locs = self.collection.get_start_locations()
        assert "t" in locs.columns
        assert locs["t"].tolist() == [
            Timestamp("2018-01-01 12:00:00"),
            Timestamp("2018-01-01 12:00:00"),
        ]

    def test_timestamp_column_present_in_end_locations(self):
        locs = self.collection.get_end_locations()
        assert "t" in locs.columns
        assert locs["t"].tolist() == [
            Timestamp("2018-01-01 14:15:00"),
            Timestamp("2018-01-02 13:15:00"),
        ]

    def test_get_end_locations(self):
        locs = self.collection.get_end_locations()
        assert len(locs) == 2
        assert locs.iloc[0].geometry in [Point(9, 9), Point(190, 10)]
        assert locs.iloc[0].id in [1, 2]
        assert locs.iloc[0].obj == "A"
        assert locs.iloc[0].val in [4, 3]
        assert locs.iloc[0].val2 in ["d", "h"]
        assert locs.iloc[1].geometry in [Point(9, 9), Point(190, 10)]
        assert locs.iloc[0].geometry != locs.iloc[1].geometry
        assert isinstance(locs, GeoDataFrame)
        assert locs.crs == CRS_METRIC

    def test_get_segments_between(self):
        collection = self.collection.get_segments_between(
            datetime(2018, 1, 1, 12, 6, 0), datetime(2018, 1, 1, 14, 10, 0)
        )
        assert len(collection) == 1
        assert collection.trajectories[0].to_linestring().wkt == "LINESTRING (6 0, 6 6)"

    def test_get_intersecting(self):
        polygon = Polygon([(-1, -1), (-1, 1), (1, 1), (1, -1), (-1, -1)])
        collection = self.collection.get_intersecting(polygon)
        assert len(collection) == 1
        assert collection.trajectories[0] == self.collection.trajectories[0]

    def test_intersection(self):
        feature = {
            "geometry": {
                "type": "Polygon",
                "coordinates": [[(-1, -1), (-1, 1), (1, 1), (1, -1), (-1, -1)]],
            },
            "properties": {"id": 1, "name": "foo"},
        }
        collection = self.collection.intersection(feature)
        assert len(collection) == 1

    def test_clip(self):
        polygon = Polygon([(-1, -1), (-1, 1), (1, 1), (1, -1), (-1, -1)])
        collection = self.collection.copy()
        collection = self.collection.clip(polygon)
        assert len(collection) == 1
        assert collection.trajectories[0].to_linestring().wkt == "LINESTRING (0 0, 1 0)"

    def test_clip_with_multipolygon(self):
        polygon = MultiPolygon(
            [
                Polygon([(-1, -1), (-1, 1), (1, 1), (1, -1), (-1, -1)]),
                Polygon([(5, 1), (7, 1), (7, 3), (5, 3), (5, 1)]),
            ]
        )
        collection = self.collection.clip(polygon)
        assert len(collection) == 2
        assert collection.trajectories[0].to_linestring().wkt == "LINESTRING (0 0, 1 0)"
        assert collection.trajectories[1].to_linestring().wkt == "LINESTRING (6 1, 6 3)"

    """ Fails, related to https://github.com/movingpandas/movingpandas/discussions/235
    def test_clip_with_multipolygon2(self):
        polygon = MultiPolygon([
            Polygon([(-1, -1), (-1, 1), (1, 1), (1, -1), (-1, -1)]),
            Polygon([(3, -1), (3, 1), (4, 1), (4, -1), (3, -1)])
        ])
        collection = self.collection.clip(polygon)
        assert len(collection) == 2
        assert collection.trajectories[0].to_linestring().wkt == "LINESTRING (0 0, 1 0)"
        assert collection.trajectories[1].to_linestring().wkt == "LINESTRING (3 0, 4 0)"
    """

    def test_clip_with_min_length(self):
        polygon = Polygon([(-1, -1), (-1, 1), (1, 1), (1, -1), (-1, -1)])
        collection = self.collection.copy()
        collection.min_length = 1
        collection = collection.clip(polygon)
        assert len(collection) == 1
        collection = self.collection.copy()
        collection.min_length = 2
        collection = collection.clip(polygon)
        assert len(collection) == 0

    def test_filter(self):
        assert len(self.collection.filter("obj", "A")) == 2
        assert len(self.collection.filter("obj", ["A"])) == 2
        assert len(self.collection.filter("obj", ["B"])) == 0
        assert len(self.collection.filter("obj", [1])) == 0

    def test_get_min_and_max(self):
        assert self.collection.get_min("val") == 2
        assert self.collection.get_max("val") == 10

    def test_plot_exists(self):
        from matplotlib.axes import Axes

        result = self.collection.plot()
        assert isinstance(result, Axes)

    @requires_folium
    def test_explore_exists(self):
        from folium.folium import Map

        if has_geopandas1:
            plot = self.collection.explore()
            assert isinstance(plot, Map)
        else:
            with pytest.raises(NotImplementedError):
                plot = self.collection.explore()

    @requires_holoviews
    def test_hvplot_exists(self):
        import holoviews

        result = self.collection_latlon.hvplot()
        assert isinstance(result, holoviews.core.overlay.Overlay)

        plot = self.collection_latlon.hvplot()
        assert isinstance(plot, holoviews.core.overlay.Overlay)
        assert len(plot.Path.ddims) == 2

        plot = self.collection_latlon.hvplot(color="red")
        assert isinstance(plot, holoviews.core.overlay.Overlay)

        plot = self.collection_latlon.hvplot(c="id")
        assert isinstance(plot, holoviews.core.overlay.Overlay)

        plot = self.collection_latlon.hvplot(c="id", colormap={1: "red", 2: "blue"})
        assert isinstance(plot, holoviews.core.overlay.Overlay)

        plot = self.collection_latlon.hvplot_pts()
        assert isinstance(plot, holoviews.core.overlay.Overlay)

        plot = self.collection_latlon.hvplot_pts(c="id")
        assert isinstance(plot, holoviews.core.overlay.Overlay)

        plot = self.collection_latlon.hvplot_pts(c="speed")
        assert isinstance(plot, holoviews.core.overlay.Overlay)

        plot = self.collection_latlon.hvplot_pts(c="id", colormap={1: "red", 2: "blue"})
        assert isinstance(plot, holoviews.core.overlay.Overlay)

    def test_plot_existing_column(self):
        from matplotlib.axes import Axes

        result = self.collection.plot(column="val")
        assert isinstance(result, Axes)

    def test_plot_speed(self):
        from matplotlib.axes import Axes

        result = self.collection.plot(column="speed")
        assert isinstance(result, Axes)

    def test_plot_speed_not_altering_collection(self):
        self.collection.plot(column="speed")
        assert all(
            [
                "speed" not in traj.df.columns.values
                for traj in self.collection.trajectories
            ]
        )

    def test_traj_with_less_than_two_points(self):
        df = pd.DataFrame(
            [[1, "A", Point(0, 0), datetime(2018, 1, 1, 12, 0, 0), 9, "a"]],
            columns=["id", "obj", "geometry", "t", "val", "val2"],
        ).set_index("t")
        geo_df = GeoDataFrame(df, crs=CRS_METRIC)
        tc = TrajectoryCollection(geo_df, "id", obj_id_col="obj")
        assert len(tc) == 0

    def test_traj_with_two_points_at_the_same_time(self):
        df = pd.DataFrame(
            [
                [1, "A", Point(0, 0), datetime(2018, 1, 1, 12, 0, 0), 9, "a"],
                [1, "A", Point(0, 0), datetime(2018, 1, 1, 12, 0, 0), 9, "a"],
            ],
            columns=["id", "obj", "geometry", "t", "val", "val2"],
        ).set_index("t")
        geo_df = GeoDataFrame(df, crs=CRS_METRIC)
        tc = TrajectoryCollection(geo_df, "id", obj_id_col="obj")
        assert len(tc) == 0

    def test_iteration(self):
        assert sum([1 for _ in self.collection]) == len(self.collection)

    def test_iteration_error(self):
        def filter_trajectory(trajectory):
            trajectory.df = trajectory.df[trajectory.df["val"] >= 7]
            return trajectory

        trajs = [filter_trajectory(traj) for traj in self.collection]

        lengths = (1, 2)
        for traj, length in zip(trajs, lengths):
            assert len(traj.df) == length

        collection = copy(self.collection)
        collection.trajectories = trajs
        with pytest.raises(ValueError):
            for _ in collection:
                pass

    def test_add_traj_id(self):
        self.collection.add_traj_id()
        result1 = self.collection.trajectories[0].df[TRAJ_ID_COL_NAME].tolist()
        assert result1 == [1, 1, 1, 1]
        result2 = self.collection.trajectories[1].df[TRAJ_ID_COL_NAME].tolist()
        assert result2 == [2, 2, 2, 2]

    def test_add_traj_id_overwrite_raises_error(self):
        gdf = self.geo_df.copy()
        gdf[TRAJ_ID_COL_NAME] = "a"
        collection = TrajectoryCollection(gdf, "id", obj_id_col="obj")
        with pytest.raises(RuntimeError):
            collection.add_traj_id()

    def test_add_speed(self):
        self.collection.add_speed()
        print(self.collection.to_point_gdf())
        result0 = self.collection.trajectories[0].df[SPEED_COL_NAME].tolist()
        assert result0[0] == pytest.approx(0.01667, 0.001)
        result1 = self.collection.trajectories[1].df[SPEED_COL_NAME].tolist()
        assert result1[0] == pytest.approx(0.01667, 0.001)
        assert len(result0) == 4

    def test_add_speed_multithreaded(self):
        self.collection.trajectories += self.collection.trajectories
        expected = self.collection.copy()
        expected.add_speed()
        self.collection.add_speed(n_threads=2)
        print(self.collection.to_point_gdf())
        result0 = self.collection.trajectories[0].df[SPEED_COL_NAME].tolist()
        assert result0 == expected.trajectories[0].df[SPEED_COL_NAME].tolist()
        result1 = self.collection.trajectories[-1].df[SPEED_COL_NAME].tolist()
        assert result1 == expected.trajectories[-1].df[SPEED_COL_NAME].tolist()
        assert len(expected.trajectories) == len(self.collection.trajectories)
        assert len(expected.trajectories) == 4

    def test_add_acceleration(self):
        self.collection.add_acceleration()
        result1 = self.collection.trajectories[0].df[ACCELERATION_COL_NAME].tolist()
        assert len(result1) == 4

    def test_add_acceleration_multithreaded(self):
        expected = self.collection.copy()
        expected.add_acceleration()
        print(expected.to_point_gdf())
        self.collection.add_acceleration(n_threads=2)
        print(self.collection.to_point_gdf())
        result0 = self.collection.trajectories[0].df[ACCELERATION_COL_NAME].tolist()
        assert result0 == expected.trajectories[0].df[ACCELERATION_COL_NAME].tolist()
        result1 = self.collection.trajectories[1].df[ACCELERATION_COL_NAME].tolist()
        assert result1 == expected.trajectories[1].df[ACCELERATION_COL_NAME].tolist()

    def test_add_direction(self):
        self.collection.add_direction()
        result = self.collection.trajectories[0].df[DIRECTION_COL_NAME].tolist()
        assert len(result) == 4

    def test_add_direction_multithreaded(self):
        expected = self.collection.copy()
        expected.add_direction()
        print(expected.to_point_gdf())
        self.collection.add_direction(n_threads=2)
        print(self.collection.to_point_gdf())
        result0 = self.collection.trajectories[0].df[DIRECTION_COL_NAME].tolist()
        assert result0 == expected.trajectories[0].df[DIRECTION_COL_NAME].tolist()
        result1 = self.collection.trajectories[1].df[DIRECTION_COL_NAME].tolist()
        assert result1 == expected.trajectories[1].df[DIRECTION_COL_NAME].tolist()

    def test_add_distance(self):
        self.collection.add_distance()
        result0 = self.collection.trajectories[0].df[DISTANCE_COL_NAME].tolist()
        assert len(result0) == 4

    def test_add_distance_multithreaded(self):
        expected = self.collection.copy()
        expected.add_distance()
        print(expected.to_point_gdf())
        self.collection.add_distance(n_threads=2)
        print(self.collection.to_point_gdf())
        result0 = self.collection.trajectories[0].df[DISTANCE_COL_NAME].tolist()
        assert result0 == expected.trajectories[0].df[DISTANCE_COL_NAME].tolist()
        result1 = self.collection.trajectories[1].df[DISTANCE_COL_NAME].tolist()
        assert result1 == expected.trajectories[1].df[DISTANCE_COL_NAME].tolist()

    def test_add_angular_difference(self):
        self.collection.add_angular_difference()
        result1 = (
            self.collection.trajectories[0].df[ANGULAR_DIFFERENCE_COL_NAME].tolist()
        )
        assert len(result1) == 4

    def test_add_angular_difference_multithreaded(self):
        expected = self.collection.copy()
        expected.add_angular_difference()
        print(expected.to_point_gdf())
        self.collection.add_angular_difference(n_threads=2)
        print(self.collection.to_point_gdf())
        result0 = (
            self.collection.trajectories[0].df[ANGULAR_DIFFERENCE_COL_NAME].tolist()
        )
        assert (
            result0 == expected.trajectories[0].df[ANGULAR_DIFFERENCE_COL_NAME].tolist()
        )
        result1 = (
            self.collection.trajectories[1].df[ANGULAR_DIFFERENCE_COL_NAME].tolist()
        )
        assert (
            result1 == expected.trajectories[1].df[ANGULAR_DIFFERENCE_COL_NAME].tolist()
        )

    def test_add_timedelta(self):
        self.collection.add_timedelta()
        result1 = self.collection.trajectories[0].df[TIMEDELTA_COL_NAME].tolist()
        assert len(result1) == 4

    def test_add_timedelta_multithreaded(self):
        expected = self.collection.copy()
        expected.add_timedelta()
        print(expected.to_point_gdf())
        self.collection.add_timedelta(n_threads=2)
        print(self.collection.to_point_gdf())
        result0 = self.collection.trajectories[0].df[TIMEDELTA_COL_NAME].tolist()
        assert result0 == expected.trajectories[0].df[TIMEDELTA_COL_NAME].tolist()
        result1 = self.collection.trajectories[1].df[TIMEDELTA_COL_NAME].tolist()
        assert result1 == expected.trajectories[1].df[TIMEDELTA_COL_NAME].tolist()

    def test_to_point_gdf(self):
        point_gdf = self.collection.to_point_gdf()
        point_gdf.to_file("temp.gpkg", layer="points", driver="GPKG")
        assert_frame_equal(point_gdf, self.geo_df)

    def test_to_line_gdf(self):
        temp_df = self.geo_df.drop(columns=["obj", "val", "val2"])
        tc = TrajectoryCollection(temp_df, "id")
        line_gdf = tc.to_line_gdf()
        line_gdf.to_file("temp.gpkg", layer="lines", driver="GPKG")
        t1 = [
            datetime(2018, 1, 1, 12, 0),
            datetime(2018, 1, 1, 12, 6),
            datetime(2018, 1, 1, 14, 10),
            datetime(2018, 1, 1, 14, 15),
        ]
        t2 = [
            datetime(2018, 1, 1, 12, 0, 0),
            datetime(2018, 1, 1, 12, 6, 0),
            datetime(2018, 1, 2, 13, 10, 0),
            datetime(2018, 1, 2, 13, 15, 0),
        ]
        df2 = pd.DataFrame(
            [
                [1, t1[1], t1[0], LineString([(0, 0), (6, 0)])],
                [1, t1[2], t1[1], LineString([(6, 0), (6, 6)])],
                [1, t1[3], t1[2], LineString([(6, 6), (9, 9)])],
                [2, t2[1], t2[0], LineString([(10, 10), (16, 10)])],
                [2, t2[2], t2[1], LineString([(16, 10), (16, 16)])],
                [2, t2[3], t2[2], LineString([(16, 16), (190, 10)])],
            ],
            columns=["id", "t", "prev_t", "geometry"],
        )
        expected_line_gdf = GeoDataFrame(df2, crs=CRS_METRIC)
        assert_frame_equal(line_gdf, expected_line_gdf)

    def test_to_traj_gdf(self):
        temp_df = self.geo_df.drop(columns=["obj", "val", "val2"])
        tc = TrajectoryCollection(temp_df, "id")
        traj_gdf = tc.to_traj_gdf()
        traj_gdf.to_file("temp.gpkg", layer="trajs", driver="GPKG")
        rows = [
            {
                "id": 1,
                "start_t": datetime(2018, 1, 1, 12, 0, 0),
                "end_t": datetime(2018, 1, 1, 14, 15, 0),
                "geometry": LineString([(0, 0), (6, 0), (6, 6), (9, 9)]),
                "length": 12 + sqrt(18),
                "direction": 45.0,
            },
            {
                "id": 2,
                "start_t": datetime(2018, 1, 1, 12, 0, 0),
                "end_t": datetime(2018, 1, 2, 13, 15, 0),
                "geometry": LineString([(10, 10), (16, 10), (16, 16), (190, 10)]),
                "length": 12 + sqrt(174 * 174 + 36),
                "direction": 90.0,
            },
        ]
        df2 = pd.DataFrame(rows)
        expected_line_gdf = GeoDataFrame(df2, crs=CRS_METRIC)

        assert_frame_equal(traj_gdf, expected_line_gdf)

    def test_to_traj_gdf_aggregate(self):
        temp_df = self.geo_df.drop(columns=["val2"])
        tc = TrajectoryCollection(temp_df, "id")
        traj_gdf = tc.to_traj_gdf(
            agg={"obj": "mode", "val": ["min", "q5", "q75", "max"]}
        )
        print(traj_gdf)
        rows = [
            {
                "id": 1,
                "start_t": datetime(2018, 1, 1, 12, 0, 0),
                "end_t": datetime(2018, 1, 1, 14, 15, 0),
                "geometry": LineString([(0, 0), (6, 0), (6, 6), (9, 9)]),
                "length": 12 + sqrt(18),
                "direction": 45.0,
                "obj_mode": "A",
                "val_min": 2,
                "val_q5": 2.3,
                "val_q75": 6.0,
                "val_max": 9,
            },
            {
                "id": 2,
                "start_t": datetime(2018, 1, 1, 12, 0, 0),
                "end_t": datetime(2018, 1, 2, 13, 15, 0),
                "geometry": LineString([(10, 10), (16, 10), (16, 16), (190, 10)]),
                "length": 12 + sqrt(174 * 174 + 36),
                "direction": 90.0,
                "obj_mode": "A",
                "val_min": 3,
                "val_q5": 3.45,
                "val_q75": 7.75,
                "val_max": 10,
            },
        ]
        df2 = pd.DataFrame(rows)
        expected_line_gdf = GeoDataFrame(df2, crs=CRS_METRIC)

        assert_frame_equal(traj_gdf, expected_line_gdf)

    def test_to_mf_json(self):
        json = self.collection.to_mf_json(datetime_to_str=False)
        assert (
            str(json)
            == """{'type': 'FeatureCollection', 'features': [{'type': 'Feature', 'properties': {'id': 1, 'obj': 'A', 'val': 9, 'val2': 'a'}, 'temporalGeometry': {'type': 'MovingPoint', 'coordinates': [(0.0, 0.0), (6.0, 0.0), (6.0, 6.0), (9.0, 9.0)], 'datetimes': [Timestamp('2018-01-01 12:00:00'), Timestamp('2018-01-01 12:06:00'), Timestamp('2018-01-01 14:10:00'), Timestamp('2018-01-01 14:15:00')]}}, {'type': 'Feature', 'properties': {'id': 2, 'obj': 'A', 'val': 10, 'val2': 'e'}, 'temporalGeometry': {'type': 'MovingPoint', 'coordinates': [(10.0, 10.0), (16.0, 10.0), (16.0, 16.0), (190.0, 10.0)], 'datetimes': [Timestamp('2018-01-01 12:00:00'), Timestamp('2018-01-01 12:06:00'), Timestamp('2018-01-02 13:10:00'), Timestamp('2018-01-02 13:15:00')]}}]}"""  # noqa F401
        )



# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from datetime import datetime
from pyproj import CRS
from geopandas import GeoDataFrame
from shapely.geometry import Point
from shapely.wkt import loads

from movingpandas.trajectory_collection import TrajectoryCollection

from .test_trajectory import make_traj, Node
from . import requires_stonesoup

CRS_METRIC = CRS.from_user_input(31256)
CRS_LATLON = CRS.from_user_input(4326)


class TestTrajectorySmoother:
    def setup_method(self):
        df = pd.DataFrame(
            [
                [1, "A", Point(0, 0), datetime(2018, 1, 1, 12, 0, 0)],
                [1, "A", Point(6, 0), datetime(2018, 1, 1, 12, 6, 0)],
                [1, "A", Point(6, 6), datetime(2018, 1, 1, 14, 10, 0)],
                [1, "A", Point(9, 9), datetime(2018, 1, 1, 14, 15, 0)],
                [2, "A", Point(10, 10), datetime(2018, 1, 1, 12, 0, 0)],
                [2, "A", Point(16, 10), datetime(2018, 1, 1, 12, 6, 0)],
                [2, "A", Point(16, 16), datetime(2018, 1, 2, 13, 10, 0)],
                [2, "A", Point(190, 19), datetime(2018, 1, 2, 13, 15, 0)],
            ],
            columns=["id", "obj", "geometry", "t"],
        ).set_index("t")
        self.df = GeoDataFrame(df, crs=CRS_METRIC)
        self.collection = TrajectoryCollection(self.df, "id", obj_id_col="obj")

    @requires_stonesoup
    def test_kalman_smoother_cv_single(self):
        from movingpandas.trajectory_smoother import KalmanSmootherCV

        traj = make_traj(
            [
                Node(),
                Node(1, 0.1, second=1),
                Node(2, 0.2, second=2),
                Node(3, 0, second=3),
                Node(3, 3, second=4),
            ]
        )
        result = KalmanSmootherCV(traj).smooth(
            process_noise_std=0.1, measurement_noise_std=10
        )
        truth = loads(
            "LINESTRING ("
            "1.7982884510469304 0.6585353696461531, "
            "1.7986579478436395 0.6588189259897214, "
            "1.7996332413712328 0.6596163790136829, "
            "1.8009712874015955 0.6608076466642573, "
            "1.8024490723366013 0.6622216786861853)"
        )
        assert np.allclose(result.to_linestring().xy, truth.xy)

    @requires_stonesoup
    def test_kalman_smoother_cv_single_tuple(self):
        from movingpandas.trajectory_smoother import KalmanSmootherCV

        traj = make_traj(
            [
                Node(),
                Node(1, 0.1, second=1),
                Node(2, 0.2, second=2),
                Node(3, 0, second=3),
                Node(3, 3, second=4),
            ]
        )
        result = KalmanSmootherCV(traj).smooth(
            process_noise_std=(0.1, 0.1), measurement_noise_std=(10, 10)
        )
        truth = loads(
            "LINESTRING ("
            "1.7982884510469304 0.6585353696461531, "
            "1.7986579478436395 0.6588189259897214, "
            "1.7996332413712328 0.6596163790136829, "
            "1.8009712874015955 0.6608076466642573, "
            "1.8024490723366013 0.6622216786861853)"
        )
        assert np.allclose(result.to_linestring().xy, truth.xy)

    @requires_stonesoup
    def test_kalman_smoother_cv_single_geo(self):
        from movingpandas.trajectory_smoother import KalmanSmootherCV

        traj = make_traj(
            [
                Node(),
                Node(1, 0.1, second=1),
                Node(2, 0.2, second=2),
                Node(3, 0, second=3),
                Node(3, 3, second=4),
            ],
            crs=CRS_LATLON,
        )
        result = KalmanSmootherCV(traj).smooth(
            process_noise_std=0.1, measurement_noise_std=10
        )
        truth = loads(
            "LINESTRING ("
            "1.7982884510469301 0.6587980686797733, "
            "1.7986579478436397 0.6590817401695466, "
            "1.7996332413712322 0.6598795189933684, "
            "1.8009712874015955 0.6610712754609851, "
            "1.8024490723366011 0.6624858873670953)"
        )
        assert np.allclose(result.to_linestring().xy, truth.xy)

    @requires_stonesoup
    def test_kalman_smoother_sv_collection(self):
        from movingpandas.trajectory_smoother import KalmanSmootherCV

        collection = KalmanSmootherCV(self.collection).smooth(
            process_noise_std=0.1, measurement_noise_std=10
        )
        truth_1 = loads(
            "LINESTRING ("
            "0.0044805333934456 0.0000761144223231, "
            "5.9954842501546315 -0.0000845436868686, "
            "6.00040594977978 6.000156500844182, "
            "8.999629266671697 8.999851928420188)"
        )
        truth_2 = loads(
            "LINESTRING ("
            "10.004429320057262 10.000008962523324, "
            "15.995566251010434 9.999990965241574, "
            "16.000875056273347 16.000014621277007, "
            "189.9991293726476 18.99998545095791)"
        )
        assert len(collection) == 2
        assert np.allclose(collection.trajectories[0].to_linestring().xy, truth_1.xy)
        assert np.allclose(collection.trajectories[1].to_linestring().xy, truth_2.xy)



import json
from typing import Callable, Dict

from geopandas import GeoDataFrame
from pandas import DataFrame


def _datetime_to_string(dt, format="%Y-%m-%d %H:%M:%S"):
    return dt.strftime(format)


def gdf_to_mf_json(
    gdf: GeoDataFrame,
    traj_id_column: str,
    datetime_column: str,
    temporal_columns: list = None,
    temporal_columns_static_fields: Dict[str, Dict] = None,
    interpolation: str = None,
    crs=None,
    trs=None,
    datetime_encoder: Callable = None,
    datetime_to_str: bool = False,
    # simplified typing hint due to https://github.com/movingpandas/movingpandas/issues/345  # noqa F401
) -> dict:
    """
    Converts a GeoDataFrame to a dictionary compatible with the Moving Features JSON
    (MF-JSON) specification.

    Args:
        gdf (GeoDataFrame): The input GeoDataFrame to convert.
        traj_id_column (str): The name of the column in the GeoDataFrame that
            represents the trajectory identifier.
        datetime_column (str): The name of the column in the GeoDataFrame that
            represents the datetime information.
        temporal_columns (list, optional): A list of column names in the GeoDataFrame
         that represent additional temporal properties. Defaults to None.
        temporal_columns_static_fields (Dict[str, Dict], optional): A dictionary mapping
            column names to static fields associated with the corresponding temporal
            property. Defaults to None.
        interpolation (str, optional): The interpolation method used for the temporal
            geometry. Defaults to None.
        crs (optional): Coordinate reference system for the MF-JSON. Defaults to None.
        trs (optional): Temporal reference system for the MF-JSON. Defaults to None.
        datetime_encoder (Callable[[any], str|int], optional): A function that encodes
            the datetime values in the GeoDataFrame to a string ( IETF RFC 3339 ) or
            integer ( Timestamp, milliseconds ). Defaults to None.
    Returns:
        dict: The MF-JSON representation of the GeoDataFrame as a dictionary.
    """

    _raise_error_if_invalid_arguments(gdf, datetime_column, traj_id_column)

    if not temporal_columns:
        temporal_columns = []

    rows = []

    if datetime_to_str:
        datetime_encoder = _datetime_to_string

    for identifier, row in gdf.groupby(traj_id_column):
        datetimes = _retrieve_datetimes_from_row(datetime_column, datetime_encoder, row)

        properties = row.drop(
            columns=[
                "geometry",
                datetime_column,
                traj_id_column,
                *temporal_columns,
            ]
        )

        if properties.empty:
            encoded_properties = {}
        else:
            encoded_properties = properties.to_dict(orient="records")[0]

        trajectory_data = {
            "type": "Feature",
            "properties": {
                traj_id_column: identifier,
                **encoded_properties,
            },
            "temporalGeometry": {
                "type": "MovingPoint",
                "coordinates": list(zip(row.geometry.x, row.geometry.y)),
                "datetimes": datetimes,
            },
        }

        if interpolation:
            trajectory_data["temporalGeometry"]["interpolation"] = interpolation

        if crs:
            trajectory_data["crs"] = crs

        if trs:
            trajectory_data["trs"] = trs

        if temporal_columns:
            temporal_properties_data = _encode_temporal_properties(
                datetimes, row, temporal_columns, temporal_columns_static_fields
            )

            trajectory_data["temporalProperties"] = [temporal_properties_data]

        # Appending each trajectory data to the list of rows
        rows.append(trajectory_data)

    return {"type": "FeatureCollection", "features": rows}


def _raise_error_if_invalid_arguments(
    gdf: GeoDataFrame, datetime_column: str, traj_id_property: str
):
    if not isinstance(gdf, GeoDataFrame):
        raise TypeError(
            "Not a GeoDataFrame, but a {} was supplied. This function only works with"
            " GeoDataFrames.".format(type(gdf))
        )
    # Check if both datetime_column and trip_id_property are in the GeoDataFrame
    if datetime_column not in gdf.columns:
        raise ValueError(
            f"The datetime_column {datetime_column} is not in the GeoDataFrame."
        )
    if traj_id_property not in gdf.columns:
        raise ValueError(
            f"The traj_id_property {traj_id_property} is not in the GeoDataFrame."
        )


def _retrieve_datetimes_from_row(datetime_column, datetime_encoder, row):
    datetimes = row[datetime_column].tolist()
    if datetime_encoder:
        datetimes = [datetime_encoder(dt) for dt in datetimes]
    return datetimes


def _encode_temporal_properties(
    datetimes, row, temporal_properties, temporal_properties_static_fields
):
    temporal_properties_data = {
        "datetimes": datetimes,
    }
    for prop in temporal_properties:
        temporal_properties_data[prop] = {
            "values": row[prop].tolist(),
        }
        if prop in (temporal_properties_static_fields or {}):
            temporal_properties_data[prop].update(
                temporal_properties_static_fields[prop]
            )
    return temporal_properties_data


def read_mf_json(json_file_path, traj_id_property=None, traj_id=0):
    """
    Reads OGC Moving Features Encoding Extension JSON files.
    MovingFeatures files are turned into Trajectory objects.
    MovingFeatureCollection files are turned into TrajectoryCollection objects.
    More info: http://www.opengis.net/doc/BP/mf-json/1.0

    Parameters
    ----------
    json_file_path : str
        Path to the JSON file
    traj_id_property : str
        Name of the MovingFeature JSON property to be used as trajectory ID
    traj_id : any
        Trajectory ID value to be used if no traj_id_property is supplied

    Returns
    -------
    Trajectory or TrajectoryCollection
    """
    with open(json_file_path, "r") as f:
        data = json.loads(f.read())
    return read_mf_dict(data, traj_id, traj_id_property)


def read_mf_dict(data: dict, traj_id=0, traj_id_property=None):
    """
    Reads OGC Moving Features Encoding Extension JSON files from a dictionary.
    MovingFeatures files are turned into Trajectory objects.
    MovingFeatureCollection files are turned into TrajectoryCollection objects.
    More info: http://www.opengis.net/doc/BP/mf-json/1.0

    data : dict
        JSON data as a dictionary
    traj_id_property : str
        Name of the MovingFeature JSON property to be used as trajectory ID
    traj_id : any
        Trajectory ID value to be used if no traj_id_property is supplied

    Returns
    """
    return _create_objects_from_mf_json_dict(data, traj_id_property, traj_id)


def _create_objects_from_mf_json_dict(data, traj_id_property=None, traj_id=0):
    if not isinstance(data, dict):
        raise ValueError("Not a supported MovingFeatures JSON")
    if data["type"] == "Feature" and "temporalGeometry" in data:
        return _create_traj_from_mfjson_movingpoint(data, traj_id_property, traj_id)
    elif data["type"] == "Feature" and "geometry" in data:
        return _create_traj_from_mfjson_trajectory(data, traj_id_property, traj_id)
    elif data["type"] == "FeatureCollection" and "features" in data:
        return _create_tc_from_mfcollection_json(data, traj_id_property)
    else:
        raise ValueError(
            f"Not a supported MovingFeatures JSON: "
            f"expected Feature or FeatureCollection "
            f"but got {data['type']}"
        )


def _get_id_property_value(data, traj_id_property):
    try:
        return data["properties"][traj_id_property]
    except KeyError:
        raise ValueError(
            f"No property '{traj_id_property}'. "
            f"Available properties are: {data['properties'].keys()}"
        )


def _create_geometry_from_linestring(data):
    if data["geometry"]["type"] == "LineString":
        t = data["properties"]["datetimes"]
        x, y = map(list, zip(*data["geometry"]["coordinates"]))
        return DataFrame(list(zip(t, x, y)), columns=["t", "x", "y"])
    else:
        raise RuntimeError(
            f"Not a supported MovingFeatures JSON: "
            f"geometry type must be LineString "
            f"(but is {data['geometry']['type']}"
        )


def _add_properties(df, data):
    # https://docs.ogc.org/is/19-045r3/19-045r3.html#_mf_json_trajectory_encoding

    for property_name, values in data["properties"].items():
        if len(values) == len(df):
            # The array of a time-varying attribute with linear interpolation
            # has elements with the same number of positions of the "datetimes"
            # member; and
            if property_name != "datetimes":
                df[property_name] = values
        elif len(values) == len(df) - 1:
            # The array of a time-varying attribute with step interpolation
            # has elements with one less than the number of positions of the
            # "datetimes" member;
            values.append(values[-1])
            df[property_name] = values
        elif len(values) == 1:
            # The array of a fixed attribute during the whole of "datetimes"
            # has only one element.
            df[property_name] = values[0]
        else:
            pass

    return df


def _create_traj_from_mfjson_trajectory(data, traj_id_property, traj_id):
    from movingpandas import Trajectory

    df = _create_geometry_from_linestring(data)

    if "properties" in data:
        df = _add_properties(df, data)

    return Trajectory(df, traj_id, t="t", x="x", y="y")


def _get_temporal_properties(data):
    cols = []
    props = []
    for key, values in data.items():
        if key == "datetimes":
            cols.append("t")
            props.append(values)
        else:
            cols.append(key)
            props.append(values["values"])
    transposed = [list(i) for i in zip(*props)]
    return DataFrame(transposed, columns=cols)


def _create_geometry_from_movingpoint(data):
    if data["temporalGeometry"]["type"] == "MovingPoint":
        t = data["temporalGeometry"]["datetimes"]
        x, y = map(list, zip(*data["temporalGeometry"]["coordinates"]))
        return DataFrame(list(zip(t, x, y)), columns=["t", "x", "y"])
    else:
        raise RuntimeError(
            f"Not a supported MovingFeatures JSON: "
            f"temporalGeometry type must be MovingPoint "
            f"(but is {data['temporalGeometry']['type']}"
        )


def _create_traj_from_mfjson_movingpoint(data, traj_id_property, traj_id):
    from movingpandas import Trajectory

    df = _create_geometry_from_movingpoint(data)
    if traj_id_property:
        traj_id = _get_id_property_value(data, traj_id_property)
    if "temporalProperties" in data:
        for property_group in data["temporalProperties"]:
            df.set_index("t", inplace=True)
            df = df.join(_get_temporal_properties(property_group).set_index("t"))
            df["t"] = df.index
    return Trajectory(df, traj_id, t="t", x="x", y="y", traj_id_col=traj_id_property)


def _create_tc_from_mfcollection_json(data, traj_id_property):
    from movingpandas import TrajectoryCollection

    assert (
        traj_id_property is not None
    ), "traj_id_property must be supplied when reading a collection of trajectories"

    trajectories = []
    for feature in data["features"]:
        trajectories.append(
            _create_traj_from_mfjson_movingpoint(feature, traj_id_property, None)
        )

    return TrajectoryCollection(trajectories)



import pytest
import importlib


def _importorskip(modname):
    try:
        importlib.import_module(modname)
        has = True
    except ImportError:
        has = False
    func = pytest.mark.skipif(not has, reason=f"requires {modname}")
    return has, func


def _skip_if_geopandas_pre_1():
    from importlib.metadata import version

    if version("geopandas") >= "1.0.0":
        has = True
    else:
        has = False
    func = pytest.mark.skipif(not has, reason="requires geopandas >= 1.0")
    return has, func


has_stonesoup, requires_stonesoup = _importorskip("stonesoup")
has_holoviews, requires_holoviews = _importorskip("holoviews")
has_folium, requires_folium = _importorskip("folium")
has_geopandas1, requires_geopandas1 = _skip_if_geopandas_pre_1()
